# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import hashlib
import os

import pytest

from pex import dist_metadata, resolver, targets
from pex.pip.tool import PackageIndexConfiguration
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.locked_resolve import LockConfiguration, LockedResolve, LockStyle
from pex.resolve.lockfile.create import LockObserver
from pex.resolve.resolved_requirement import Pin
from pex.resolve.resolver_configuration import PipConfiguration
from pex.resolve.testing import normalize_locked_resolve
from pex.resolver import Downloaded, LocalDistribution, WheelBuilder
from pex.typing import TYPE_CHECKING
from pex.util import CacheHelper

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Tuple


def normalize(
    locked_resolves,  # type: Tuple[LockedResolve, ...]
    skip_additional_artifacts=False,  # type: bool
    skip_urls=False,  # type: bool
    skip_verified=False,  # type: bool
):
    # type: (...) -> Tuple[LockedResolve, ...]
    return tuple(
        normalize_locked_resolve(
            lock,
            skip_additional_artifacts=skip_additional_artifacts,
            skip_urls=skip_urls,
            skip_verified=skip_verified,
        )
        for lock in locked_resolves
    )


def create_lock_observer(lock_configuration):
    # type: (LockConfiguration) -> LockObserver
    pip_configuration = PipConfiguration()
    package_index_configuration = PackageIndexConfiguration.create(
        resolver_version=pip_configuration.resolver_version,
        indexes=pip_configuration.repos_configuration.indexes,
        find_links=pip_configuration.repos_configuration.find_links,
        network_configuration=pip_configuration.network_configuration,
    )
    return LockObserver(
        root_requirements=(),
        lock_configuration=lock_configuration,
        resolver=ConfiguredResolver(pip_configuration=pip_configuration),
        wheel_builder=WheelBuilder(
            package_index_configuration,
            prefer_older_binary=pip_configuration.prefer_older_binary,
            use_pep517=pip_configuration.use_pep517,
            build_isolation=pip_configuration.build_isolation,
        ),
        package_index_configuration=package_index_configuration,
    )


def create_lock(
    lock_configuration,  # type: LockConfiguration
    **kwargs  # type: Any
):
    # type: (...) -> Tuple[Downloaded, Tuple[LockedResolve, ...]]
    lock_observer = create_lock_observer(lock_configuration)
    downloaded = resolver.download(observer=lock_observer, **kwargs)
    return downloaded, lock_observer.lock(downloaded)


@pytest.mark.parametrize(
    "requirements",
    (
        pytest.param(["ansicolors==1.1.8"], id="pinned-no-transitive-deps"),
        pytest.param(["isort==4.3.21"], id="pinned-transitive-deps"),
        pytest.param(["ansicolors"], id="float-no-transitive-deps"),
        pytest.param(["isort"], id="float-transitive-deps"),
    ),
)
@pytest.mark.parametrize(
    "lock_configuration",
    (
        pytest.param(LockConfiguration(style=LockStyle.STRICT), id="strict"),
        pytest.param(LockConfiguration(style=LockStyle.SOURCES), id="sources"),
    ),
)
def test_lock_single_target(
    tmpdir,  # type: Any
    requirements,  # type: Iterable[str]
    lock_configuration,  # type: LockConfiguration
):
    # type: (...) -> None

    downloaded, locked_resolves = create_lock(lock_configuration, requirements=requirements)
    assert 1 == len(locked_resolves)
    lock = locked_resolves[0]

    assert targets.current().platform.tag == lock.platform_tag

    def pin(local_distribution):
        # type: (LocalDistribution) -> Pin
        project_name_and_version = dist_metadata.project_name_and_version(local_distribution.path)
        assert project_name_and_version is not None
        return Pin.canonicalize(project_name_and_version)

    local_distributions_by_pin = {
        pin(local_dist): local_dist for local_dist in downloaded.local_distributions
    }  # type: Dict[Pin, LocalDistribution]

    assert sorted(local_distributions_by_pin) == sorted(
        locked_req.pin for locked_req in lock.locked_requirements
    ), (
        "Expected the actual set of downloaded distributions to match the set of pinned "
        "requirements in the lock."
    )

    for locked_req in lock.locked_requirements:
        fingerprint = locked_req.artifact.fingerprint
        assert fingerprint.hash == CacheHelper.hash(
            path=local_distributions_by_pin[locked_req.pin].path,
            hasher=lambda: hashlib.new(fingerprint.algorithm),
        ), (
            "Expected the fingerprint of the downloaded distribution to match the fingerprint "
            "recorded in the lock."
        )

    find_links_repo = os.path.join(str(tmpdir), "find-links")
    os.mkdir(find_links_repo)
    for local_dist in downloaded.local_distributions:
        os.symlink(
            local_dist.path, os.path.join(find_links_repo, os.path.basename(local_dist.path))
        )
    _, find_links_locked_resolves = create_lock(
        lock_configuration,
        requirements=requirements,
        indexes=[],
        find_links=[find_links_repo],
    )
    assert normalize(
        locked_resolves, skip_additional_artifacts=True, skip_urls=True, skip_verified=True
    ) == normalize(
        find_links_locked_resolves,
        skip_additional_artifacts=True,
        skip_urls=True,
        skip_verified=True,
    ), (
        "Expected a find-links lock to match an equivalent PyPI lock except for the primary "
        "artifact urls and their verification status and lack of additional artifacts (since these "
        "are never downloaded; but instead, just recorded)."
    )



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import functools
import json
import logging
import os
import subprocess
import sys
import tempfile
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace, _ActionsContainer
from contextlib import contextmanager

from pex import pex_warnings
from pex.argparse import HandleBoolAction
from pex.common import safe_mkdtemp, safe_open
from pex.result import Error, Ok, Result
from pex.typing import TYPE_CHECKING, Generic, cast
from pex.variables import ENV, Variables
from pex.version import __version__

if TYPE_CHECKING:
    from typing import (
        IO,
        Any,
        Dict,
        Iterable,
        Iterator,
        NoReturn,
        Optional,
        Sequence,
        Type,
        TypeVar,
    )

    import attr  # vendor:skip
else:
    from pex.third_party import attr

if TYPE_CHECKING:
    _T = TypeVar("_T")


def try_run_program(
    program,  # type: str
    args,  # type: Iterable[str]
    url=None,  # type: Optional[str]
    error=None,  # type: Optional[str]
    **kwargs  # type: Any
):
    # type: (...) -> Result
    try:
        subprocess.check_call([program] + list(args), **kwargs)
        return Ok()
    except OSError as e:
        msg = [error] if error else []
        msg.append("Do you have `{}` installed on the $PATH?: {}".format(program, e))
        if url:
            msg.append(
                "Find more information on `{program}` at {url}.".format(program=program, url=url)
            )
        return Error("\n".join(msg))
    except subprocess.CalledProcessError as e:
        return Error(str(e), exit_code=e.returncode)


def try_open_file(
    path,  # type: str
    error=None,  # type: Optional[str]
):
    # type: (...) -> Result
    opener, url = (
        ("xdg-open", "https://www.freedesktop.org/wiki/Software/xdg-utils/")
        if "Linux" == os.uname()[0]
        else ("open", None)
    )
    with open(os.devnull, "wb") as devnull:
        return try_run_program(opener, [path], url=url, error=error, stdout=devnull)


@attr.s(frozen=True)
class Command(object):
    @staticmethod
    def show_help(
        parser,  # type: ArgumentParser
        *_args,  # type: Any
        **_kwargs  # type: Any
    ):
        # type: (...) -> NoReturn
        parser.error("a subcommand is required")

    @staticmethod
    def register_global_arguments(
        parser,  # type: _ActionsContainer
        include_verbosity=True,  # type: bool
    ):
        # type: (...) -> None
        register_global_arguments(parser, include_verbosity=include_verbosity)

    @classmethod
    def name(cls):
        # type: () -> str
        return cls.__name__.lower()

    @classmethod
    def description(cls):
        # type: () -> Optional[str]
        return cls.__doc__

    @classmethod
    def add_arguments(cls, parser):
        # type: (ArgumentParser) -> None
        pass

    options = attr.ib()  # type: Namespace


class OutputMixin(object):
    @staticmethod
    def add_output_option(
        parser,  # type: _ActionsContainer
        entity,  # type: str
    ):
        # type: (...) -> None
        parser.add_argument(
            "-o",
            "--output",
            metavar="PATH",
            help=(
                "A file to output the {entity} to; STDOUT by default or when `-` is "
                "specified.".format(entity=entity)
            ),
        )

    @staticmethod
    def is_stdout(options):
        # type: (Namespace) -> bool
        return options.output == "-" or not options.output

    @classmethod
    @contextmanager
    def output(
        cls,
        options,  # type: Namespace
        binary=False,  # type: bool
    ):
        # type: (...) -> Iterator[IO]
        if cls.is_stdout(options):
            stdout = getattr(sys.stdout, "buffer", sys.stdout) if binary else sys.stdout
            yield stdout
        else:
            with safe_open(options.output, mode="wb" if binary else "w") as out:
                yield out


class JsonMixin(object):
    @staticmethod
    def add_json_options(
        parser,  # type: _ActionsContainer
        entity,  # type: str
        include_switch=True,  # type: bool
    ):
        flags = ("-i", "--indent") if include_switch else ("--indent",)
        parser.add_argument(
            *flags,
            type=int,
            default=None,
            help="Pretty-print {entity} json with the given indent.".format(entity=entity)
        )

    @staticmethod
    def dump_json(
        options,  # type: Namespace
        data,  # type: Dict[str, Any]
        out,  # type: IO
        **json_dump_kwargs  # type: Any
    ):
        if options.indent is not None and options.indent > 0:
            # Python 2.7 uses ', ' for the list item separator regardless of indent which is
            # different from Python 3 and leads to trailing whitespace in the output; so, we
            # normalize here to the Python 3 style for consistent, more generally useful output.
            json_dump_kwargs.update(separators=(",", ": "))
        json.dump(data, out, indent=options.indent, **json_dump_kwargs)


def register_global_arguments(
    parser,  # type: _ActionsContainer
    include_verbosity=True,  # type: bool
):
    # type: (...) -> None
    """Register Pex global environment configuration options with the given parser.

    :param parser: The parser to register global options with.
    :param include_verbosity: Whether to include the verbosity option `-v`.
    """

    group = parser.add_argument_group(title="Global options")
    if include_verbosity:
        group.add_argument(
            "-v",
            dest="verbosity",
            action="count",
            default=0,
            help="Turn on logging verbosity, may be specified multiple times.",
        )
    group.add_argument(
        "--emit-warnings",
        "--no-emit-warnings",
        dest="emit_warnings",
        action=HandleBoolAction,
        default=True,
        help=(
            "Emit runtime UserWarnings on stderr. If false, only emit them when PEX_VERBOSE "
            "is set."
        ),
    )
    group.add_argument(
        "--pex-root",
        dest="pex_root",
        default=None,
        help=(
            "Specify the pex root used in this invocation of pex "
            "(if unspecified, uses {}).".format(ENV.PEX_ROOT)
        ),
    )
    group.add_argument(
        "--disable-cache",
        dest="disable_cache",
        default=False,
        action="store_true",
        help="Disable caching in the pex tool entirely.",
    )

    group.add_argument(
        "--cache-dir",
        dest="cache_dir",
        default=None,
        help=(
            "DEPRECATED: Use --pex-root instead. The local cache directory to use for speeding up "
            "requirement lookups."
        ),
    )
    group.add_argument(
        "--tmpdir",
        dest="tmpdir",
        default=tempfile.gettempdir(),
        help="Specify the temporary directory Pex and its subprocesses should use.",
    )
    group.add_argument(
        "--rcfile",
        dest="rc_file",
        default=None,
        help=(
            "An additional path to a pexrc file to read during configuration parsing, in addition "
            "to reading `/etc/pexrc` and `~/.pexrc`. If `PEX_IGNORE_RCFILES=true`, then all rc "
            "files will be ignored."
        ),
    )


class GlobalConfigurationError(Exception):
    """Indicates an error processing global options."""


@contextmanager
def _configured_env(options):
    # type: (Namespace) -> Iterator[None]
    if options.rc_file or not ENV.PEX_IGNORE_RCFILES:
        with ENV.patch(**Variables(rc=options.rc_file).copy()):
            yield
    else:
        yield


@contextmanager
def global_environment(options):
    # type: (Namespace) -> Iterator[Dict[str, str]]
    """Configures the Pex global environment.

    This includes configuration of basic Pex infrastructure like logging, warnings and the
    `PEX_ROOT` to use.

    :param options: The global options registered by `register_global_arguments`.
    :yields: The configured global environment.
    :raises: :class:`GlobalConfigurationError` if invalid global option values were specified.
    """
    if not hasattr(options, "rc_file"):
        # We don't register the global args on the root command (but do on every subcommand).
        # So if the user runs just `pex` with no subcommand we must not attempt to use those
        # global args, including rc_file, which we check for here as a representative of the
        # global args.
        # Note that we can't use command_type here because the legacy command line parser in
        # pex/bin/pex.py uses this function as well, and it doesn't set command_type.
        with ENV.patch() as env:
            yield env
    with _configured_env(options):
        verbosity = Variables.PEX_VERBOSE.strip_default(ENV)
        if verbosity is None:
            verbosity = getattr(options, "verbosity", 0)

        emit_warnings = True
        if not options.emit_warnings:
            emit_warnings = False
        if emit_warnings and ENV.PEX_EMIT_WARNINGS is not None:
            emit_warnings = ENV.PEX_EMIT_WARNINGS

        with ENV.patch(PEX_VERBOSE=str(verbosity), PEX_EMIT_WARNINGS=str(emit_warnings)):
            pex_warnings.configure_warnings(env=ENV)

            # Ensure the TMPDIR is an absolute path (So subprocesses that change CWD can find it)
            # and that it exists.
            tmpdir = os.path.realpath(options.tmpdir)
            if not os.path.exists(tmpdir):
                raise GlobalConfigurationError(
                    "The specified --tmpdir does not exist: {}".format(tmpdir)
                )
            if not os.path.isdir(tmpdir):
                raise GlobalConfigurationError(
                    "The specified --tmpdir is not a directory: {}".format(tmpdir)
                )
            tempfile.tempdir = os.environ["TMPDIR"] = tmpdir

            if options.cache_dir:
                pex_warnings.warn("The --cache-dir option is deprecated, use --pex-root instead.")
                if options.pex_root and options.cache_dir != options.pex_root:
                    raise GlobalConfigurationError(
                        "Both --cache-dir and --pex-root were passed with conflicting values. "
                        "Just set --pex-root."
                    )

            if options.disable_cache:

                def warn_ignore_pex_root(set_via):
                    pex_warnings.warn(
                        "The pex root has been set via {via} but --disable-cache is also set. "
                        "Ignoring {via} and disabling caches.".format(via=set_via)
                    )

                if options.cache_dir:
                    warn_ignore_pex_root("--cache-dir")
                elif options.pex_root:
                    warn_ignore_pex_root("--pex-root")
                elif os.environ.get("PEX_ROOT"):
                    warn_ignore_pex_root("PEX_ROOT")

                pex_root = safe_mkdtemp()
            else:
                pex_root = options.cache_dir or options.pex_root or ENV.PEX_ROOT

            with ENV.patch(PEX_ROOT=pex_root, TMPDIR=tmpdir) as env:
                yield env


if TYPE_CHECKING:
    _C = TypeVar("_C", bound=Command)


class Main(Generic["_C"]):
    def __init__(
        self,
        command_types,  # type: Iterable[Type[_C]]
        description=None,  # type: Optional[str]
        subparsers_description=None,  # type: Optional[str]
        prog=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self._prog = prog
        self._description = description or self.__doc__
        self._subparsers_description = subparsers_description
        self._command_types = command_types

    def add_arguments(self, parser):
        # type: (ArgumentParser) -> None
        pass

    @contextmanager
    def parsed_command(self, args=None):
        # type: (Optional[Sequence[str]]) -> Iterator[_C]
        logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)

        # By default, let argparse derive prog from sys.argv[0].
        prog = self._prog
        if os.path.basename(sys.argv[0]) == "__main__.py":
            prog = "{python} -m {module}".format(
                python=sys.executable, module=".".join(type(self).__module__.split(".")[:-1])
            )

        parser = ArgumentParser(
            prog=prog,
            formatter_class=ArgumentDefaultsHelpFormatter,
            description=self._description,
        )
        parser.add_argument("-V", "--version", action="version", version=__version__)
        parser.set_defaults(command_type=functools.partial(Command.show_help, parser))
        self.add_arguments(parser)
        if self._command_types:
            subparsers = parser.add_subparsers(description=self._subparsers_description)
            for command_type in self._command_types:
                name = command_type.name()
                description = command_type.description()
                help_text = description.splitlines()[0] if description else None
                command_parser = subparsers.add_parser(
                    name,
                    formatter_class=ArgumentDefaultsHelpFormatter,
                    help=help_text,
                    description=description,
                )
                command_type.add_arguments(command_parser)
                command_parser.set_defaults(command_type=command_type)

        options = parser.parse_args(args=args)
        with global_environment(options):
            command_type = cast("Type[_C]", options.command_type)
            yield command_type(options)



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.enum import Enum


class BinPath(Enum["BinPath.Value"]):
    class Value(Enum.Value):
        pass

    FALSE = Value("false")
    PREPEND = Value("prepend")
    APPEND = Value("append")



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import warnings

import pytest

from pex import pex_warnings
from pex.common import temporary_dir
from pex.compatibility import PY2
from pex.pex_warnings import PEXWarning
from pex.testing import environment_as
from pex.typing import TYPE_CHECKING
from pex.util import named_temporary_file
from pex.variables import NoValueError, Variables

if TYPE_CHECKING:
    from typing import Any


def test_process_pydoc():
    # type: () -> None
    def thing():
        # no pydoc
        pass

    assert Variables.process_pydoc(thing.__doc__) == ("Unknown", "Unknown")

    def other_thing():
        """Type

        Properly
                formatted
            text.
        """

    assert Variables.process_pydoc(other_thing.__doc__) == ("Type", "Properly formatted text.")


def test_iter_help():
    # type: () -> None
    for variable_name, variable_type, variable_text in Variables.iter_help():
        assert variable_name.startswith("PEX_")
        assert "\n" not in variable_type
        assert "\n" not in variable_text


def test_pex_bool_variables():
    # type: () -> None
    assert Variables(environ={})._maybe_get_bool("NOT_HERE") is None
    with pytest.raises(NoValueError):
        Variables(environ={})._get_bool("NOT_HERE")

    for value in ("0", "faLsE", "false"):
        assert Variables(environ={"HERE": value})._get_bool("HERE") is False
    for value in ("1", "TrUe", "true"):
        assert Variables(environ={"HERE": value})._get_bool("HERE") is True
    with pytest.raises(ValueError):
        Variables(environ={"HERE": "garbage"})._get_bool("HERE")

    # end to end
    assert Variables().PEX_ALWAYS_CACHE is False
    assert Variables({"PEX_ALWAYS_CACHE": "1"}).PEX_ALWAYS_CACHE is True


def test_pex_string_variables():
    # type: () -> None
    assert Variables(environ={})._maybe_get_string("NOT_HERE") is None
    with pytest.raises(NoValueError):
        Variables(environ={})._get_string("NOT_HERE")
    assert Variables(environ={"HERE": "stuff"})._get_string("HERE") == "stuff"


def test_pex_get_int():
    # type: () -> None
    with pytest.raises(NoValueError):
        Variables()._get_int("HELLO")
    assert Variables(environ={"HELLO": "23"})._get_int("HELLO") == 23
    with pytest.raises(SystemExit):
        assert Variables(environ={"HELLO": "welp"})._get_int("HELLO")


def test_disable():
    v = Variables(
        environ=dict(
            PEX_DISABLE_VARIABLES="1",
            PEX_VERBOSE="42",
            PEX_PATH="other.pex:more.pex",
            PEX_SCRIPT="other",
        )
    )
    assert v.PEX_DISABLE_VARIABLES is True
    assert 0 == v.PEX_VERBOSE
    assert () == v.PEX_PATH
    assert v.PEX_SCRIPT is None
    assert v.PEX_EMIT_WARNINGS is None


def test_disable_patch():
    v = Variables(environ=dict(PEX_VERBOSE="42", PEX_PATH="other.pex:more.pex", PEX_SCRIPT="other"))
    assert v.PEX_DISABLE_VARIABLES is False
    assert 42 == v.PEX_VERBOSE
    assert ("other.pex", "more.pex") == v.PEX_PATH
    assert "other" == v.PEX_SCRIPT
    assert v.PEX_EMIT_WARNINGS is None

    with v.patch(PEX_DISABLE_VARIABLES="1", PEX_EMIT_WARNINGS="1"):
        assert v.PEX_DISABLE_VARIABLES is True
        assert 0 == v.PEX_VERBOSE
        assert () == v.PEX_PATH
        assert v.PEX_SCRIPT is None
        assert v.PEX_EMIT_WARNINGS is None


def assert_pex_vars_hermetic():
    # type: () -> None
    v = Variables()
    assert os.environ.copy() == v.copy()

    existing = os.environ.get("TEST")
    expected = (existing or "") + "different"
    assert expected != existing

    with environment_as(TEST=expected):
        assert expected != v.copy().get("TEST")


def test_pex_vars_hermetic_no_pexrc():
    # type: () -> None
    assert_pex_vars_hermetic()


def test_pex_vars_hermetic():
    # type: () -> None
    with environment_as(PEX_IGNORE_RCFILES="True"):
        assert_pex_vars_hermetic()


def test_pex_vars_disabled_hermetic():
    # type: () -> None
    with environment_as(PEX_DISABLE_VARIABLES="True"):
        assert_pex_vars_hermetic()


def test_pex_get_kv():
    # type: () -> None
    v = Variables(environ={})
    assert v._get_kv("HELLO") is None
    assert v._get_kv("=42") is None
    assert v._get_kv("TOO=MANY=COOKS") is None
    assert v._get_kv("THIS=WORKS") == ["THIS", "WORKS"]


def test_pex_from_rc():
    # type: () -> None
    with named_temporary_file(mode="w") as pexrc:
        pexrc.write("HELLO=42")
        pexrc.flush()
        v = Variables(rc=pexrc.name)
        assert v._get_int("HELLO") == 42


def test_pexrc_precedence():
    # type: () -> None
    with named_temporary_file(mode="w") as pexrc:
        pexrc.write("HELLO=FORTYTWO")
        pexrc.flush()
        v = Variables(rc=pexrc.name, environ={"HELLO": "42"})
        assert v._get_int("HELLO") == 42


def test_rc_ignore():
    # type: () -> None
    with named_temporary_file(mode="w") as pexrc:
        pexrc.write("HELLO=FORTYTWO")
        pexrc.flush()
        v = Variables(rc=pexrc.name, environ={"PEX_IGNORE_RCFILES": "True"})
        assert "HELLO" not in v._environ


def test_pex_vars_defaults_stripped():
    # type: () -> None
    v = Variables(environ={})

    # bool
    assert v.PEX_ALWAYS_CACHE is not None
    assert Variables.PEX_ALWAYS_CACHE.strip_default(v) is None

    # string
    assert v.PEX_PROFILE_SORT is not None
    assert Variables.PEX_PROFILE_SORT.strip_default(v) is None

    # int
    assert v.PEX_VERBOSE is not None
    assert Variables.PEX_VERBOSE.strip_default(v) is None


def test_pex_root_unwriteable():
    # type: () -> None
    with temporary_dir() as td:
        pex_root = os.path.realpath(os.path.join(td, "pex_root"))
        os.mkdir(pex_root, 0o444)

        env = Variables(environ=dict(PEX_ROOT=pex_root))

        with warnings.catch_warnings(record=True) as log:
            assert pex_root != env.PEX_ROOT

        assert 1 == len(log)
        message = log[0].message
        assert isinstance(message, PEXWarning)
        assert pex_root in str(message)
        assert env.PEX_ROOT is not None
        assert env.PEX_ROOT in str(message)

        assert (
            env.PEX_ROOT == env.PEX_ROOT
        ), "When an ephemeral PEX_ROOT is materialized it should be stable."


def test_pex_vars_value_or(tmpdir):
    # type: (Any) -> None
    v = Variables(environ={})

    assert v.PEX_ROOT is not None, "Expected PEX_ROOT to be a defaulted variable."

    pex_root = str(tmpdir)
    assert pex_root == Variables.PEX_ROOT.value_or(v, pex_root)

    unwriteable_pex_root = os.path.join(pex_root, "unwriteable")
    os.mkdir(unwriteable_pex_root, 0o444)
    assert unwriteable_pex_root != Variables.PEX_ROOT.value_or(v, unwriteable_pex_root), (
        "Expected the fallback to be validated, and in the case of PEX_ROOT, replaced with a "
        "writeable tmp dir"
    )


def test_patch():
    # type: () -> None
    v = Variables(environ=dict(PEX_VERBOSE="3", PEX_PYTHON="jython", PEX_EMIT_WARNINGS="True"))
    assert v.PEX_VERBOSE == 3
    assert v.PEX_PYTHON == "jython"
    assert v.PEX_EMIT_WARNINGS is True
    assert v.PEX_FORCE_LOCAL is False

    with v.patch(PEX_VERBOSE="1", PEX_EMIT_WARNINGS=None, PEX_FORCE_LOCAL="True") as env:
        assert env["PEX_VERBOSE"] == "1"
        assert env["PEX_PYTHON"] == "jython"
        assert "PEX_EMIT_WARNINGS" not in env
        assert env["PEX_FORCE_LOCAL"] == "True"

        assert v.PEX_VERBOSE == 1
        assert v.PEX_PYTHON == "jython"
        assert v.PEX_EMIT_WARNINGS is None

        # If the assertion is flipped from `is True` to `is False` this test fails; so MyPy is just
        # confused here about the statement being unreachable.
        assert v.PEX_FORCE_LOCAL is True  # type: ignore[unreachable]


@pytest.mark.skipif(
    PY2,
    reason=(
        "The `warnings.catch_warnings` mechanism doesn't work properly under CPython 2.7 & pypy2 "
        "across multiple tests. Since we only use `warnings.catch_warnings` in unit tests and "
        "the mechanisms tested here are also tested in integration tests under CPython 2.7 & pypy "
        "we accept that these unit tests appear un-fixable without a lot of warnings mocking."
    ),
)
def test_warnings():
    # type: () -> None
    environ = dict(
        PEX_IGNORE_ERRORS="true",
        PEX_ALWAYS_CACHE="true",
        PEX_FORCE_LOCAL="true",
        PEX_UNZIP="true",
    )
    with warnings.catch_warnings(record=True) as events:
        pex_warnings.configure_warnings(Variables(environ={}))
        env = Variables(environ=environ)
    assert env.PEX_IGNORE_ERRORS is True
    assert env.PEX_ALWAYS_CACHE is True
    assert env.PEX_FORCE_LOCAL is True
    assert env.PEX_UNZIP is True

    warning_by_message_first_sentence = {
        str(event.message).split(". ")[0]: event.message for event in events
    }
    assert all(
        isinstance(warning, PEXWarning) for warning in warning_by_message_first_sentence.values()
    )
    assert tuple(
        sorted(
            "The `{}` env var is deprecated".format(env_var)
            for env_var in ("PEX_ALWAYS_CACHE", "PEX_FORCE_LOCAL", "PEX_UNZIP")
        )
    ) == tuple(sorted(warning_by_message_first_sentence))


def test_empty_pex_path_issue_1936():
    # type: () -> None

    assert () == Variables(environ={}).PEX_PATH
    assert () == Variables(environ={"PEX_PATH": ""}).PEX_PATH
    assert (".",) == Variables(environ={"PEX_PATH": "."}).PEX_PATH



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys

from pex.cli.testing import run_pex3
from pex.resolve.lockfile import json_codec
from pex.testing import PY38, ensure_python_interpreter, make_env
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_os_name_spoofing(tmpdir):
    # type: (Any) -> None

    lock = os.path.join(str(tmpdir), "lock.json")

    create_lock_args = [
        "lock",
        "create",
        "--style",
        "universal",
        "--resolver-version",
        "pip-2020-resolver",
        "pywinpty==2.0.6; os_name == 'nt'",
        "--interpreter-constraint",
        "CPython<4,>=3.7.5",
        "--indent",
        "2",
        "-o",
        lock,
        "--target-system",
        "linux",
        "--target-system",
        "mac",
    ]

    # This ensures that, even if the machine has a functioning cargo / rust toolchain on the PATH,
    # it becomes non-functioning without altering the system permanently. This is needed to ensure
    # we can't build the pywinpty sdist to extract needed resolve metadata from it.
    #
    # For more info on this corner to reproducing the prior failure, see:
    #   https://github.com/pantsbuild/pex/issues/1856#issuecomment-1193054493
    env = make_env(RUSTC=os.devnull)

    # The above attempt to get pywinpty dependency metadata by building the sdist requires a
    # CPython>=3.7.5 on the PATH which we arrange for if not already present here.
    if sys.version_info[:3] < (3, 7, 5):
        python_path = os.environ["PATH"].split(os.pathsep) + [ensure_python_interpreter(PY38)]
        create_lock_args.extend(["--python-path", os.pathsep.join(python_path)])

    run_pex3(*create_lock_args, env=env).assert_success()

    lockfile = json_codec.load(lockfile_path=lock)
    assert 1 == len(lockfile.locked_resolves)

    locked_resolve = lockfile.locked_resolves[0]
    assert 0 == len(locked_resolve.locked_requirements), (
        "Since we're not running on `os_name == 'nt'`, we shouldn't have been able to lock any "
        "requirements at all."
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
import sys

import pytest

from pex.testing import IS_LINUX, PY38, ensure_python_interpreter, make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(not IS_LINUX, reason="We only release from Linux in CI.")
def test_packaging(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None
    pex = os.path.join(str(tmpdir), "pex.pex")
    package_script = os.path.join(pex_project_dir, "scripts", "package.py")
    run_pex_command(
        args=[
            "toml",
            pex_project_dir,
            "--",
            package_script,
            "-v",
            "--pex-output-file",
            pex,
        ],
        # The package script requires Python 3.
        python=sys.executable if sys.version_info[0] >= 3 else ensure_python_interpreter(PY38),
    ).assert_success()
    assert os.path.exists(pex), "Expected {pex} to be created by {package_script}.".format(
        pex=pex, package_script=package_script
    )
    # The packaged Pex PEX should work with all Pythons we support, including the current test
    # interpreter.
    subprocess.check_call(args=[sys.executable, pex, "-V"], env=make_env(PEX_PYTHON=sys.executable))



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import pytest

from pex import testing
from pex.interpreter import PythonInterpreter
from pex.platforms import Platform
from pex.testing import PY27, PY38, PY39, PY310, ensure_python_interpreter


@pytest.fixture(scope="session")
def pex_project_dir():
    # type: () -> str
    return testing.pex_project_dir()


@pytest.fixture
def current_interpreter():
    # type: () -> PythonInterpreter
    return PythonInterpreter.get()


@pytest.fixture
def current_platform(current_interpreter):
    # type: (PythonInterpreter) -> Platform
    return current_interpreter.platform


@pytest.fixture
def py27():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY27))


@pytest.fixture
def py38():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY38))


@pytest.fixture
def py39():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY39))


@pytest.fixture
def py310():
    # type: () -> PythonInterpreter
    return PythonInterpreter.from_binary(ensure_python_interpreter(PY310))



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os
import subprocess
import sys
from textwrap import dedent

from pex import third_party
from pex.build_system import DEFAULT_BUILD_BACKEND
from pex.build_system.pep_518 import BuildSystem, load_build_system
from pex.common import safe_mkdtemp
from pex.dist_metadata import DistMetadata, Distribution
from pex.jobs import Job, SpawnedJob
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.resolvers import Resolver
from pex.result import Error, try_
from pex.targets import Target, Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast
from pex.util import named_temporary_file

if TYPE_CHECKING:
    from typing import Any, Dict, Iterable, Mapping, Optional, Text, Tuple, Union

_DEFAULT_BUILD_SYSTEMS = {}  # type: Dict[PipVersionValue, BuildSystem]


def _default_build_system(
    pip_version,  # type: PipVersionValue
    target,  # type: Target
    resolver,  # type: Resolver
):
    # type: (...) -> BuildSystem
    global _DEFAULT_BUILD_SYSTEMS
    build_system = _DEFAULT_BUILD_SYSTEMS.get(pip_version)
    if build_system is None:
        with TRACER.timed(
            "Building {build_backend} build_backend PEX".format(build_backend=DEFAULT_BUILD_BACKEND)
        ):
            extra_env = {}  # type: Dict[str, str]
            if pip_version is PipVersion.VENDORED:
                requires = ["setuptools", "wheel"]
                resolved = tuple(
                    Distribution.load(dist_location)
                    for dist_location in third_party.expose(requires)
                )
                extra_env.update(__PEX_UNVENDORED__="1")
            else:
                requires = [pip_version.setuptools_requirement, pip_version.wheel_requirement]
                resolved = tuple(
                    installed_distribution.fingerprinted_distribution.distribution
                    for installed_distribution in resolver.resolve_requirements(
                        requirements=requires,
                        targets=Targets.from_target(target),
                    ).installed_distributions
                )
            build_system = try_(
                BuildSystem.create(
                    interpreter=target.get_interpreter(),
                    requires=requires,
                    resolved=resolved,
                    build_backend=DEFAULT_BUILD_BACKEND,
                    backend_path=(),
                    **extra_env
                )
            )
            _DEFAULT_BUILD_SYSTEMS[pip_version] = build_system
    return build_system


def _get_build_system(
    pip_version,  # type: PipVersionValue
    target,  # type: Target
    resolver,  # type: Resolver
    project_directory,  # type: str
    extra_requirements=None,  # type: Optional[Iterable[str]]
):
    # type: (...) -> Union[BuildSystem, Error]
    custom_build_system_or_error = load_build_system(
        target, resolver, project_directory, extra_requirements=extra_requirements
    )
    if custom_build_system_or_error:
        return custom_build_system_or_error
    return _default_build_system(pip_version, target, resolver)


# Exit code 75 is EX_TEMPFAIL defined in /usr/include/sysexits.h
# this seems an appropriate signal of DNE vs execute and fail.
_HOOK_UNAVAILABLE_EXIT_CODE = 75


def is_hook_unavailable_error(error):
    # type: (Job.Error) -> bool
    return error.exitcode == _HOOK_UNAVAILABLE_EXIT_CODE


def _invoke_build_hook(
    project_directory,  # type: str
    pip_version,  # type: PipVersionValue
    target,  # type: Target
    resolver,  # type: Resolver
    hook_method,  # type: str
    hook_args=(),  # type: Iterable[Any]
    hook_extra_requirements=None,  # type: Optional[Iterable[str]]
    hook_kwargs=None,  # type: Optional[Mapping[str, Any]]
    stdout=None,  # type: Optional[int]
    stderr=None,  # type: Optional[int]
):
    # type: (...) -> Union[SpawnedJob[Any], Error]

    if not os.path.exists(project_directory):
        return Error(
            "Project directory {project_directory} does not exist.".format(
                project_directory=project_directory
            )
        )
    if not os.path.isdir(project_directory):
        return Error(
            "Project directory {project_directory} is not a directory.".format(
                project_directory=project_directory
            )
        )

    build_system_or_error = _get_build_system(
        pip_version, target, resolver, project_directory, extra_requirements=hook_extra_requirements
    )
    if isinstance(build_system_or_error, Error):
        return build_system_or_error
    build_system = build_system_or_error

    # The interfaces are spec'd here: https://peps.python.org/pep-0517
    build_backend_module, _, _ = build_system.build_backend.partition(":")
    build_backend_object = build_system.build_backend.replace(":", ".")
    with named_temporary_file(mode="r") as fp:
        args = build_system.venv_pex.execute_args(
            "-c",
            dedent(
                """\
                import json
                import sys

                import {build_backend_module}


                if not hasattr({build_backend_object}, {hook_method!r}):
                    sys.exit({hook_unavailable_exit_code})

                result = {build_backend_object}.{hook_method}(*{hook_args!r}, **{hook_kwargs!r})
                with open({result_file!r}, "w") as fp:
                    json.dump(result, fp)
                """
            ).format(
                build_backend_module=build_backend_module,
                build_backend_object=build_backend_object,
                hook_method=hook_method,
                hook_args=tuple(hook_args),
                hook_kwargs=dict(hook_kwargs) if hook_kwargs else {},
                hook_unavailable_exit_code=_HOOK_UNAVAILABLE_EXIT_CODE,
                result_file=fp.name,
            ),
        )
        process = subprocess.Popen(
            args=args,
            env=build_system.env,
            cwd=project_directory,
            stdout=stdout if stdout is not None else sys.stderr.fileno(),
            stderr=stderr if stderr is not None else sys.stderr.fileno(),
        )
        return SpawnedJob.file(
            Job(command=args, process=process),
            output_file=fp.name,
            result_func=lambda file_content: json.loads(file_content.decode("utf-8")),
        )


def build_sdist(
    project_directory,  # type: str
    dist_dir,  # type: str
    pip_version,  # type: PipVersionValue
    target,  # type: Target
    resolver,  # type: Resolver
):
    # type: (...) -> Union[Text, Error]

    extra_requirements = []
    spawned_job_or_error = _invoke_build_hook(
        project_directory,
        pip_version,
        target,
        resolver,
        hook_method="get_requires_for_build_sdist",
    )
    if isinstance(spawned_job_or_error, Error):
        return spawned_job_or_error
    try:
        extra_requirements.extend(spawned_job_or_error.await_result())
    except Job.Error as e:
        if e.exitcode != _HOOK_UNAVAILABLE_EXIT_CODE:
            return Error(
                "Failed to prepare build backend for building an sdist for local project "
                "{project_directory}: {err}\n"
                "{stderr}".format(project_directory=project_directory, err=e, stderr=e.stderr)
            )

    spawned_job_or_error = _invoke_build_hook(
        project_directory,
        pip_version,
        target,
        resolver,
        hook_method="build_sdist",
        hook_args=[dist_dir],
        hook_extra_requirements=extra_requirements,
    )
    if isinstance(spawned_job_or_error, Error):
        return spawned_job_or_error
    try:
        sdist_relpath = cast(str, spawned_job_or_error.await_result())
    except Job.Error as e:
        return Error(
            "Failed to build sdist for local project {project_directory}: {err}\n"
            "{stderr}".format(project_directory=project_directory, err=e, stderr=e.stderr)
        )
    return os.path.join(dist_dir, sdist_relpath)


def spawn_prepare_metadata(
    project_directory,  # type: str
    pip_version,  # type: PipVersionValue
    target,  # type: Target
    resolver,  # type: Resolver
):
    # type: (...) -> SpawnedJob[DistMetadata]

    extra_requirements = []
    spawned_job = try_(
        _invoke_build_hook(
            project_directory,
            pip_version,
            target,
            resolver,
            hook_method="get_requires_for_build_wheel",
        )
    )
    try:
        extra_requirements.extend(spawned_job.await_result())
    except Job.Error as e:
        if e.exitcode != _HOOK_UNAVAILABLE_EXIT_CODE:
            raise e

    build_dir = os.path.join(safe_mkdtemp(), "build")
    os.mkdir(build_dir)
    spawned_job = try_(
        _invoke_build_hook(
            project_directory,
            pip_version,
            target,
            resolver,
            hook_method="prepare_metadata_for_build_wheel",
            hook_args=[build_dir],
            hook_extra_requirements=extra_requirements,
        )
    )
    return spawned_job.map(lambda _: DistMetadata.load(build_dir))



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import itertools
import json
import os
import re
from collections import OrderedDict, defaultdict

from pex import hashing
from pex.common import safe_mkdtemp
from pex.compatibility import urlparse
from pex.dist_metadata import ProjectNameAndVersion, Requirement
from pex.hashing import Sha256
from pex.orderedset import OrderedSet
from pex.pep_440 import Version
from pex.pip import foreign_platform
from pex.pip.download_observer import Patch, PatchSet
from pex.pip.local_project import digest_local_project
from pex.pip.log_analyzer import LogAnalyzer
from pex.pip.vcs import fingerprint_downloaded_vcs_archive
from pex.pip.version import PipVersionValue
from pex.requirements import ArchiveScheme, VCSRequirement, VCSScheme
from pex.resolve.locked_resolve import LockConfiguration, LockStyle, TargetSystem
from pex.resolve.pep_691.fingerprint_service import FingerprintService
from pex.resolve.pep_691.model import Endpoint
from pex.resolve.resolved_requirement import (
    ArtifactURL,
    Fingerprint,
    PartialArtifact,
    Pin,
    ResolvedRequirement,
)
from pex.resolve.resolvers import Resolver
from pex.targets import Target
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import (
        DefaultDict,
        Dict,
        Iterable,
        List,
        Mapping,
        Optional,
        Pattern,
        Set,
        Text,
        Tuple,
    )

    import attr  # vendor:skip

    from pex.requirements import ParsedRequirement
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class LockResult(object):
    resolved_requirements = attr.ib()  # type: Tuple[ResolvedRequirement, ...]
    local_projects = attr.ib()  # type: Tuple[str, ...]


@attr.s(frozen=True)
class Credentials(object):
    username = attr.ib()  # type: str
    password = attr.ib(default=None)  # type: Optional[str]

    def are_redacted(self):
        # type: () -> bool

        # N.B.: Pip redacts here: pex/vendor/_vendored/pip/pip/_internal/utils/misc.py
        return "****" in (self.username, self.password)

    def render_basic_auth(self):
        # type: () -> str
        if self.password is not None:
            return "{username}:{password}".format(username=self.username, password=self.password)
        return self.username


@attr.s(frozen=True)
class Netloc(object):
    host = attr.ib()  # type: str
    port = attr.ib(default=None)  # type: Optional[int]

    def render_host_port(self):
        # type: () -> str
        if self.port is not None:
            return "{host}:{port}".format(host=self.host, port=self.port)
        return self.host


@attr.s(frozen=True)
class CredentialedURL(object):
    @classmethod
    def parse(cls, url):
        # type: (Text) -> CredentialedURL

        url_info = urlparse.urlparse(url)

        # The netloc component of the parsed url combines username, password, host and port. We need
        # to track username and password; so we break up netloc into its four components.
        credentials = (
            Credentials(username=url_info.username, password=url_info.password)
            if url_info.username
            else None
        )

        netloc = Netloc(host=url_info.hostname, port=url_info.port) if url_info.hostname else None

        return cls(
            scheme=url_info.scheme,
            credentials=credentials,
            netloc=netloc,
            path=url_info.path,
            params=url_info.params,
            query=url_info.query,
            fragment=url_info.fragment,
        )

    scheme = attr.ib()  # type: str
    credentials = attr.ib()  # type: Optional[Credentials]
    netloc = attr.ib()  # type: Optional[Netloc]
    path = attr.ib()  # type: str
    params = attr.ib()  # type: str
    query = attr.ib()  # type: str
    fragment = attr.ib()  # type: str

    @property
    def has_redacted_credentials(self):
        # type: () -> bool
        if self.credentials is None:
            return False
        return self.credentials.are_redacted()

    def strip_credentials(self):
        # type: () -> CredentialedURL
        return attr.evolve(self, credentials=None)

    def strip_params_query_and_fragment(self):
        # type: () -> CredentialedURL
        return attr.evolve(self, params="", query="", fragment="")

    def inject_credentials(self, credentials):
        # type: (Optional[Credentials]) -> CredentialedURL
        return attr.evolve(self, credentials=credentials)

    def __str__(self):
        # type: () -> str

        netloc = ""
        if self.netloc is not None:
            host_port = self.netloc.render_host_port()
            netloc = (
                "{credentials}@{host_port}".format(
                    credentials=self.credentials.render_basic_auth(), host_port=host_port
                )
                if self.credentials
                else host_port
            )

        return urlparse.urlunparse(
            (self.scheme, netloc, self.path, self.params, self.query, self.fragment)
        )


@attr.s(frozen=True)
class VCSURLManager(object):
    @staticmethod
    def _normalize_vcs_url(credentialed_url):
        # type: (CredentialedURL) -> str
        return str(credentialed_url.strip_credentials().strip_params_query_and_fragment())

    @classmethod
    def create(cls, requirements):
        # type: (Iterable[ParsedRequirement]) -> VCSURLManager

        credentials_by_normalized_url = {}  # type: Dict[str, Optional[Credentials]]
        for requirement in requirements:
            if not isinstance(requirement, VCSRequirement):
                continue
            credentialed_url = CredentialedURL.parse(requirement.url)
            vcs_url = "{vcs}+{url}".format(
                vcs=requirement.vcs, url=cls._normalize_vcs_url(credentialed_url)
            )
            credentials_by_normalized_url[vcs_url] = credentialed_url.credentials
        return cls(credentials_by_normalized_url)

    _credentials_by_normalized_url = attr.ib()  # type: Mapping[str, Optional[Credentials]]

    def normalize_url(self, url):
        # type: (str) -> str

        credentialed_url = CredentialedURL.parse(url)
        if credentialed_url.has_redacted_credentials:
            normalized_vcs_url = self._normalize_vcs_url(credentialed_url)
            credentials = self._credentials_by_normalized_url.get(normalized_vcs_url)
            if credentials is not None:
                credentialed_url = credentialed_url.inject_credentials(credentials)
        return str(credentialed_url)


class AnalyzeError(Exception):
    """Indicates an error analyzing lock data."""


@attr.s(frozen=True)
class ArtifactBuildResult(object):
    url = attr.ib()  # type: ArtifactURL
    pin = attr.ib()  # type: Pin
    requirement = attr.ib()  # type: Requirement


@attr.s(frozen=True)
class ArtifactBuildObserver(object):
    _done_building_pattern = attr.ib()  # type: Pattern
    _artifact_url = attr.ib()  # type: ArtifactURL

    def is_done_building(self, line):
        # type: (str) -> bool
        return self._done_building_pattern.search(line) is not None

    def build_result(self, line):
        # type: (str) -> Optional[ArtifactBuildResult]

        match = re.search(
            r"Source in .+ has version (?P<version>[^\s]+), which satisfies requirement "
            r"(?P<requirement>.+) .*from {url}".format(url=re.escape(self._artifact_url.raw_url)),
            line,
        )
        if not match:
            return None

        version = Version(match.group("version"))
        requirement = Requirement.parse(match.group("requirement"))
        pin = Pin(project_name=requirement.project_name, version=version)
        return ArtifactBuildResult(
            url=self._artifact_url,
            pin=pin,
            requirement=requirement,
        )


class Locker(LogAnalyzer):
    def __init__(
        self,
        target,  # type: Target
        root_requirements,  # type: Iterable[ParsedRequirement]
        pip_version,  # type: PipVersionValue
        resolver,  # type: Resolver
        lock_configuration,  # type: LockConfiguration
        download_dir,  # type: str
        fingerprint_service=None,  # type: Optional[FingerprintService]
    ):
        # type: (...) -> None

        self._target = target
        self._vcs_url_manager = VCSURLManager.create(root_requirements)
        self._pip_version = pip_version
        self._resolver = resolver
        self._lock_configuration = lock_configuration
        self._download_dir = download_dir
        self._fingerprint_service = fingerprint_service or FingerprintService()

        self._saved = set()  # type: Set[Pin]
        self._selected_path_to_pin = {}  # type: Dict[str, Pin]

        self._resolved_requirements = []  # type: List[ResolvedRequirement]
        self._pep_691_endpoints = set()  # type: Set[Endpoint]
        self._links = defaultdict(
            OrderedDict
        )  # type: DefaultDict[Pin, OrderedDict[ArtifactURL, PartialArtifact]]
        self._artifact_build_observer = None  # type: Optional[ArtifactBuildObserver]
        self._local_projects = OrderedSet()  # type: OrderedSet[str]
        self._lock_result = None  # type: Optional[LockResult]

    @property
    def style(self):
        # type: () -> LockStyle.Value
        return self._lock_configuration.style

    @property
    def requires_python(self):
        # type: () -> Tuple[str, ...]
        return self._lock_configuration.requires_python

    def should_collect(self, returncode):
        # type: (int) -> bool
        return returncode == 0

    @staticmethod
    def _extract_resolve_data(artifact_url):
        # type: (ArtifactURL) -> Tuple[Pin, PartialArtifact]

        fingerprint = None  # type: Optional[Fingerprint]
        fingerprint_match = re.search(
            r"[^#]+#(?P<algorithm>[^=]+)=(?P<hash>.*)$", artifact_url.raw_url
        )
        if fingerprint_match:
            algorithm = fingerprint_match.group("algorithm")
            hash_ = fingerprint_match.group("hash")
            fingerprint = Fingerprint(algorithm=algorithm, hash=hash_)

        pin = Pin.canonicalize(ProjectNameAndVersion.from_filename(artifact_url.path))
        partial_artifact = PartialArtifact(artifact_url, fingerprint)
        return pin, partial_artifact

    def analyze(self, line):
        # type: (str) -> LogAnalyzer.Continue[None]

        # The log sequence for processing a resolved requirement is as follows (log lines irrelevant
        # to our purposes omitted):
        #
        #   1.)   "... Found link <url1> ..."
        #   ...
        #   1.)   "... Found link <urlN> ..."
        #   2.)   "... Added <varying info ...> to build tracker ..."
        # * 3.)   Lines related to extracting metadata from <requirement> if the selected
        #         distribution is an sdist in any form (VCS, local directory, source archive).
        # * 3.5.) "... Source in <tmp> has version <version>, which satisfies requirement "
        #         "<requirement> from <url> ..."
        #   4.)   "... Removed <requirement> from <url> ... from build tracker ..."
        #   5.)   "... Saved <download dir>/<artifact file>

        # The lines in section 3 can contain this same pattern of lines if the metadata extraction
        # proceeds via PEP-517 which recursively uses Pip to resolve build dependencies. We want to
        # ignore this recursion since a lock should only contain install requirements and not build
        # requirements (If a build proceeds differently tomorrow than today then we don't care as
        # long as the final built artifact hashes the same. In other words, we completely rely on a
        # cryptographic fingerprint for reproducibility and security guarantees from a lock).

        # The section 4 line will be present for requirements that represent either local source
        # directories or VCS requirements and can be used to learn their version.

        if self._artifact_build_observer:
            if self._artifact_build_observer.is_done_building(line):
                self._artifact_build_observer = None
                return self.Continue()

            build_result = self._artifact_build_observer.build_result(line)
            if build_result:
                artifact_url = build_result.url
                source_fingerprint = None  # type: Optional[Fingerprint]
                verified = False
                if isinstance(artifact_url.scheme, VCSScheme):
                    source_fingerprint, archive_path = fingerprint_downloaded_vcs_archive(
                        download_dir=self._download_dir,
                        project_name=str(build_result.pin.project_name),
                        version=str(build_result.pin.version),
                        vcs=artifact_url.scheme.vcs,
                    )
                    verified = True
                    selected_path = os.path.basename(archive_path)
                    artifact_url = ArtifactURL.parse(
                        self._vcs_url_manager.normalize_url(artifact_url.raw_url)
                    )
                    self._selected_path_to_pin[selected_path] = build_result.pin
                elif isinstance(artifact_url.scheme, ArchiveScheme.Value):
                    selected_path = os.path.basename(artifact_url.path)
                    source_archive_path = os.path.join(self._download_dir, selected_path)
                    # If Pip resolves the artifact from its own cache, we will not find it in the
                    # download dir for this run; so guard against that. In this case the existing
                    # machinery that finalizes a locks missing fingerprints will download the
                    # artifact and hash it.
                    if os.path.isfile(source_archive_path):
                        digest = Sha256()
                        hashing.file_hash(source_archive_path, digest)
                        source_fingerprint = Fingerprint.from_digest(digest)
                        verified = True
                    self._selected_path_to_pin[selected_path] = build_result.pin
                elif "file" == artifact_url.scheme:
                    digest = Sha256()
                    if os.path.isfile(artifact_url.path):
                        hashing.file_hash(artifact_url.path, digest)
                        self._selected_path_to_pin[
                            os.path.basename(artifact_url.path)
                        ] = build_result.pin
                    else:
                        digest_local_project(
                            directory=artifact_url.path,
                            digest=digest,
                            pip_version=self._pip_version,
                            target=self._target,
                            resolver=self._resolver,
                        )
                        self._local_projects.add(artifact_url.path)
                        self._saved.add(build_result.pin)
                    source_fingerprint = Fingerprint.from_digest(digest)
                    verified = True
                else:
                    raise AnalyzeError(
                        "Unexpected scheme {scheme!r} for artifact at {url}".format(
                            scheme=artifact_url.scheme, url=artifact_url
                        )
                    )

                additional_artifacts = self._links[build_result.pin]
                additional_artifacts.pop(artifact_url, None)

                self._resolved_requirements.append(
                    ResolvedRequirement(
                        requirement=build_result.requirement,
                        pin=build_result.pin,
                        artifact=PartialArtifact(
                            url=artifact_url, fingerprint=source_fingerprint, verified=verified
                        ),
                        additional_artifacts=tuple(additional_artifacts.values()),
                    )
                )
            return self.Continue()

        match = re.search(
            r"Fetched page (?P<index_url>[^\s]+) as (?P<content_type>{content_types})".format(
                content_types="|".join(
                    re.escape(content_type) for content_type in self._fingerprint_service.accept
                )
            ),
            line,
        )
        if match:
            self._pep_691_endpoints.add(
                Endpoint(url=match.group("index_url"), content_type=match.group("content_type"))
            )
            return self.Continue()

        match = re.search(
            r"Added (?P<requirement>.+) from (?P<url>[^\s]+) .*to build tracker",
            line,
        )
        if match:
            raw_requirement = match.group("requirement")
            url = ArtifactURL.parse(match.group("url"))
            if url.is_wheel:
                requirement = Requirement.parse(raw_requirement)
                pin, partial_artifact = self._extract_resolve_data(url)

                additional_artifacts = self._links[pin]
                additional_artifacts.pop(url, None)
                self._resolved_requirements.append(
                    ResolvedRequirement(
                        requirement=requirement,
                        pin=pin,
                        artifact=partial_artifact,
                        additional_artifacts=tuple(additional_artifacts.values()),
                    )
                )
                self._selected_path_to_pin[os.path.basename(url.path)] = pin
            else:
                self._artifact_build_observer = ArtifactBuildObserver(
                    done_building_pattern=re.compile(
                        r"Removed {requirement} from {url} (?:.* )?from build tracker".format(
                            requirement=re.escape(raw_requirement), url=re.escape(url.raw_url)
                        )
                    ),
                    artifact_url=url,
                )
            return self.Continue()

        match = re.search(r"Added (?P<file_url>file:.+) to build tracker", line)
        if match:
            file_url = match.group("file_url")
            self._artifact_build_observer = ArtifactBuildObserver(
                done_building_pattern=re.compile(
                    r"Removed .+ from {file_url} from build tracker".format(
                        file_url=re.escape(file_url)
                    )
                ),
                artifact_url=ArtifactURL.parse(file_url),
            )
            return self.Continue()

        match = re.search(r"Saved (?P<file_path>.+)$", line)
        if match:
            saved_path = match.group("file_path")
            self._saved.add(self._selected_path_to_pin[os.path.basename(saved_path)])
            return self.Continue()

        if self.style in (LockStyle.SOURCES, LockStyle.UNIVERSAL):
            match = re.search(r"Found link (?P<url>[^\s]+)(?: \(from .*\))?, version: ", line)
            if match:
                url = ArtifactURL.parse(match.group("url"))
                pin, partial_artifact = self._extract_resolve_data(url)
                self._links[pin][url] = partial_artifact
                return self.Continue()

        return self.Continue()

    def analysis_completed(self):
        # type: () -> None
        resolved_requirements = [
            resolved_requirement
            for resolved_requirement in self._resolved_requirements
            if resolved_requirement.pin in self._saved
        ]

        fingerprinted_artifacts = {
            artifact.url: artifact
            for artifact in self._fingerprint_service.fingerprint(
                endpoints=self._pep_691_endpoints,
                artifacts=tuple(
                    artifact
                    for resolved_requirement in resolved_requirements
                    for artifact in resolved_requirement.iter_artifacts()
                ),
            )
        }

        def maybe_fill_in_fingerprints(resolved_requirement):
            # type: (ResolvedRequirement) -> ResolvedRequirement
            return attr.evolve(
                resolved_requirement,
                artifact=fingerprinted_artifacts.get(resolved_requirement.artifact.url),
                additional_artifacts=tuple(
                    fingerprinted_artifacts.get(artifact.url)
                    for artifact in resolved_requirement.additional_artifacts
                ),
            )

        self._lock_result = LockResult(
            resolved_requirements=tuple(
                maybe_fill_in_fingerprints(resolved_requirement)
                for resolved_requirement in resolved_requirements
            ),
            local_projects=tuple(self._local_projects),
        )

    @property
    def lock_result(self):
        # type: () -> LockResult
        assert (
            self._lock_result is not None
        ), "Lock result was retrieved before analysis was complete."
        return self._lock_result


# See https://www.python.org/dev/peps/pep-0508/#environment-markers for more about these values.
_OS_NAME = {
    TargetSystem.LINUX: "posix",
    TargetSystem.MAC: "posix",
    TargetSystem.WINDOWS: "nt",
}
_PLATFORM_SYSTEM = {
    TargetSystem.LINUX: "Linux",
    TargetSystem.MAC: "Darwin",
    TargetSystem.WINDOWS: "Windows",
}
_SYS_PLATFORMS = {
    TargetSystem.LINUX: ("linux", "linux2"),
    TargetSystem.MAC: ("darwin",),
    TargetSystem.WINDOWS: ("win32",),
}

# See: https://peps.python.org/pep-0425/#platform-tag for more about the wheel platform tag.
_PLATFORM_TAG_REGEXP = {
    TargetSystem.LINUX: r"linux",
    TargetSystem.MAC: r"macosx",
    TargetSystem.WINDOWS: r"win",
}


def patch(lock_configuration):
    # type: (LockConfiguration) -> PatchSet

    if lock_configuration.style != LockStyle.UNIVERSAL:
        return PatchSet()

    patches_dir = safe_mkdtemp()
    patches = []
    if lock_configuration.requires_python:
        patches.append(
            foreign_platform.patch_requires_python(
                requires_python=lock_configuration.requires_python, patches_dir=patches_dir
            )
        )

    env = {}  # type: Dict[str, str]
    if lock_configuration.target_systems and set(lock_configuration.target_systems) != set(
        TargetSystem.values()
    ):
        target_systems = {
            "os_names": [
                _OS_NAME[target_system] for target_system in lock_configuration.target_systems
            ],
            "platform_systems": [
                _PLATFORM_SYSTEM[target_system]
                for target_system in lock_configuration.target_systems
            ],
            "sys_platforms": list(
                itertools.chain.from_iterable(
                    _SYS_PLATFORMS[target_system]
                    for target_system in lock_configuration.target_systems
                )
            ),
            "platform_tag_regexps": [
                _PLATFORM_TAG_REGEXP[target_system]
                for target_system in lock_configuration.target_systems
            ],
        }
        with open(os.path.join(patches_dir, "target_systems.json"), "w") as fp:
            json.dump(target_systems, fp)
        env.update(_PEX_TARGET_SYSTEMS_FILE=fp.name)

    patches.append(Patch.from_code_resource(__name__, "locker_patches.py", **env))

    return PatchSet(patches=tuple(patches))



# Copyright 2019 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import subprocess
from textwrap import dedent

import pytest

from pex.interpreter import spawn_python_job
from pex.jobs import _ABSOLUTE_MAX_JOBS, DEFAULT_MAX_JOBS, Job, SpawnedJob, _sanitize_max_jobs
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Any, Dict


def test_sanitize_max_jobs_none():
    # type: () -> None
    assert DEFAULT_MAX_JOBS == _sanitize_max_jobs(None)


def test_sanitize_max_jobs_less_then_one():
    # type: () -> None
    assert DEFAULT_MAX_JOBS == _sanitize_max_jobs(0)
    assert DEFAULT_MAX_JOBS == _sanitize_max_jobs(-1)
    assert DEFAULT_MAX_JOBS == _sanitize_max_jobs(-5)


def test_sanitize_max_jobs_nominal():
    # type: () -> None
    assert 1 == _sanitize_max_jobs(1)


def test_sanitize_max_jobs_too_large():
    # type: () -> None
    assert _ABSOLUTE_MAX_JOBS == _sanitize_max_jobs(_ABSOLUTE_MAX_JOBS)
    assert _ABSOLUTE_MAX_JOBS == _sanitize_max_jobs(_ABSOLUTE_MAX_JOBS + 1)
    assert _ABSOLUTE_MAX_JOBS == _sanitize_max_jobs(_ABSOLUTE_MAX_JOBS + 5)


def create_error_job(exit_code):
    # type: (int) -> Job
    return spawn_python_job(args=["-c", "import sys; sys.exit({})".format(exit_code)])


def test_spawn_wait():
    # type: () -> None
    result = object()
    assert (
        result is SpawnedJob.wait(job=spawn_python_job(args=["-V"]), result=result).await_result()
    )

    spawned_job = SpawnedJob.wait(job=create_error_job(42), result=1 / 137)
    with pytest.raises(Job.Error) as exec_info:
        spawned_job.await_result()
    assert 42 == exec_info.value.exitcode


def test_spawn_and_then(tmpdir):
    # type: (Any) -> None
    side_effect_file = os.path.join(str(tmpdir), "side.effect")

    def observe_side_effect():
        # type: () -> Dict[str, int]
        with open(side_effect_file) as fp:
            return cast("Dict[str, int]", json.load(fp))

    assert (
        {"exit_code": 42}
        == SpawnedJob.and_then(
            job=spawn_python_job(
                args=[
                    "-c",
                    dedent(
                        """\
                    import json

                    with open({side_effect_file!r}, "w") as fp:
                        json.dump({{"exit_code": 42}}, fp)
                    """
                    ).format(side_effect_file=side_effect_file),
                ]
            ),
            result_func=observe_side_effect,
        ).await_result()
    )

    spawned_job = SpawnedJob.and_then(job=create_error_job(3), result_func=lambda: 1 / 137)
    with pytest.raises(Job.Error) as exec_info:
        spawned_job.await_result()
    assert 3 == exec_info.value.exitcode


def test_spawn_stdout():
    # type: () -> None
    assert (
        "Jane\n"
        == SpawnedJob.stdout(
            job=spawn_python_job(
                args=["-c", "import sys; print(sys.stdin.read())"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            ),
            result_func=lambda stdout: stdout.decode("utf-8"),
            input=b"Jane",
        ).await_result()
    )

    spawned_job = SpawnedJob.stdout(create_error_job(137), lambda output: 42)
    with pytest.raises(Job.Error) as exec_info:
        spawned_job.await_result()
    assert 137 == exec_info.value.exitcode



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import warnings

import pytest

from pex.common import temporary_dir
from pex.inherit_path import InheritPath
from pex.interpreter_constraints import InterpreterConstraints
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.pex_warnings import PEXWarning
from pex.typing import TYPE_CHECKING
from pex.variables import Variables
from pex.version import __version__ as pex_version

if TYPE_CHECKING:
    from typing import Dict, List, Text


def test_backwards_incompatible_pex_info():
    # type: () -> None

    def make_pex_info(requirements):
        # type: (List[Text]) -> PexInfo
        return PexInfo(info={"requirements": requirements})

    # forwards compatibility
    pi = make_pex_info(["hello"])
    assert pi.requirements == OrderedSet(["hello"])

    pi = make_pex_info(["hello==0.1", "world==0.2"])
    assert pi.requirements == OrderedSet(["hello==0.1", "world==0.2"])

    # malformed
    with pytest.raises(ValueError):
        make_pex_info("hello")  # type: ignore[arg-type]

    with pytest.raises(ValueError):
        make_pex_info([("hello", False)])  # type: ignore[list-item]

    # backwards compatibility
    pi = make_pex_info(
        [
            ["hello==0.1", False, None],  # type: ignore[list-item]
            ["world==0.2", False, None],  # type: ignore[list-item]
        ]
    )
    assert pi.requirements == OrderedSet(["hello==0.1", "world==0.2"])


def assert_same_info(expected, actual):
    # type: (PexInfo, PexInfo) -> None
    assert expected.dump() == actual.dump()


def test_from_empty_env():
    # type: () -> None
    environ = Variables(environ={})
    info = {}  # type: Dict
    assert_same_info(PexInfo(info=info), PexInfo.from_env(env=environ))


def test_from_env():
    # type: () -> None
    with temporary_dir() as td:
        pex_root = os.path.realpath(os.path.join(td, "pex_root"))
        environ = dict(
            PEX_ROOT=pex_root,
            PEX_MODULE="entry:point",
            PEX_SCRIPT="script.sh",
            PEX_INHERIT_PATH="prefer",
            PEX_IGNORE_ERRORS="true",
            # These are deprecated and should be ignored by PexInfo.
            PEX_ALWAYS_CACHE="true",
            PEX_FORCE_LOCAL="true",
            PEX_UNZIP="true",
        )
        with warnings.catch_warnings(record=True):
            env = Variables(environ=environ)

        info = dict(
            pex_root=pex_root,
            entry_point="entry:point",
            script="script.sh",
            inherit_path=True,
            ignore_errors=True,
        )

    assert_same_info(PexInfo(info=info), PexInfo.from_env(env=env))


def test_build_properties():
    # type: () -> None
    assert pex_version == PexInfo.default().build_properties["pex_version"]


def test_pex_root_set_unwriteable():
    # type: () -> None
    with temporary_dir() as td:
        pex_root = os.path.realpath(os.path.join(td, "pex_root"))
        os.mkdir(pex_root, 0o444)

        pex_info = PexInfo.default()
        pex_info.pex_root = pex_root

        with warnings.catch_warnings(record=True) as log:
            assert pex_root != pex_info.pex_root

        assert 1 == len(log)
        message = log[0].message
        assert isinstance(message, PEXWarning)
        assert pex_root in str(message)
        assert pex_info.pex_root in str(message)


def test_copy():
    # type: () -> None
    default_info = PexInfo.default()
    default_info_copy = default_info.copy()
    assert default_info is not default_info_copy
    assert default_info.dump() == default_info_copy.dump()

    info = PexInfo.default()
    info.code_hash = "foo"
    info.inherit_path = InheritPath.FALLBACK
    info.add_requirement("bar==1")
    info.add_requirement("baz==2")
    info.add_distribution("bar.whl", "bar-sha")
    info.add_distribution("baz.whl", "baz-sha")
    info.interpreter_constraints = InterpreterConstraints.parse(">=2.7.18", "CPython==2.7.9")
    info_copy = info.copy()

    assert "foo" == info_copy.code_hash
    assert InheritPath.FALLBACK == info_copy.inherit_path
    assert OrderedSet(["bar==1", "baz==2"]) == info_copy.requirements
    assert {"bar.whl": "bar-sha", "baz.whl": "baz-sha"} == info_copy.distributions
    assert (
        InterpreterConstraints.parse(">=2.7.18", "CPython==2.7.9")
        == info_copy.interpreter_constraints
    )
    assert info.dump() == info_copy.dump()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys

import pytest

from pex.cli.testing import run_pex3
from pex.testing import IS_PYPY, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    IS_PYPY or sys.version_info[:3] < (3, 6, 2),
    reason="The black 22.3.0 distribution requires Python >= 3.6.2",
)
def test_lock_black(tmpdir):
    # type: (Any) -> None

    lock = os.path.join(str(tmpdir), "lock")
    lock_create_args = (
        "lock",
        "create",
        "--resolver-version",
        "pip-2020-resolver",
        "--style",
        "universal",
        "black==22.3.0",
        "-o",
        lock,
    )

    def assert_lock(*extra_lock_args, **extra_popen_args):
        run_pex3(*(lock_create_args + extra_lock_args), **extra_popen_args).assert_success()
        result = run_pex_command(args=["--lock", lock, "-c", "black", "--", "--version"])
        result.assert_success()
        assert " 22.3.0 " in result.output

    assert_lock()

    cwd = os.path.join(str(tmpdir), "cwd")
    tmpdir = os.path.join(cwd, ".tmp")
    os.makedirs(tmpdir)
    assert_lock("--tmpdir", ".tmp", cwd=cwd)



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.lock_resolver import resolve_from_lock
from pex.resolve.pex_repository_resolver import resolve_from_pex
from pex.resolve.requirement_configuration import RequirementConfiguration
from pex.resolve.resolver_configuration import (
    LockRepositoryConfiguration,
    PexRepositoryConfiguration,
)
from pex.resolve.resolvers import Installed
from pex.resolver import resolve as resolve_via_pip
from pex.result import try_
from pex.targets import Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pex.resolve.resolver_options import ResolverConfiguration


def resolve(
    targets,  # type: Targets
    requirement_configuration,  # type: RequirementConfiguration
    resolver_configuration,  # type: ResolverConfiguration
    compile_pyc=False,  # type: bool
    ignore_errors=False,  # type: bool
):
    # type: (...) -> Installed
    if isinstance(resolver_configuration, LockRepositoryConfiguration):
        lock = try_(resolver_configuration.parse_lock())
        with TRACER.timed(
            "Resolving requirements from lock file {lock_file}".format(lock_file=lock.source)
        ):
            pip_configuration = resolver_configuration.pip_configuration
            return try_(
                resolve_from_lock(
                    targets=targets,
                    lock=lock,
                    resolver=ConfiguredResolver(pip_configuration=pip_configuration),
                    requirements=requirement_configuration.requirements,
                    requirement_files=requirement_configuration.requirement_files,
                    constraint_files=requirement_configuration.constraint_files,
                    transitive=pip_configuration.transitive,
                    indexes=pip_configuration.repos_configuration.indexes,
                    find_links=pip_configuration.repos_configuration.find_links,
                    resolver_version=pip_configuration.resolver_version,
                    network_configuration=pip_configuration.network_configuration,
                    password_entries=pip_configuration.repos_configuration.password_entries,
                    build=pip_configuration.allow_builds,
                    use_wheel=pip_configuration.allow_wheels,
                    prefer_older_binary=pip_configuration.prefer_older_binary,
                    use_pep517=pip_configuration.use_pep517,
                    build_isolation=pip_configuration.build_isolation,
                    compile=compile_pyc,
                    max_parallel_jobs=pip_configuration.max_jobs,
                    pip_version=lock.pip_version,
                )
            )
    elif isinstance(resolver_configuration, PexRepositoryConfiguration):
        with TRACER.timed(
            "Resolving requirements from PEX {pex_repository}.".format(
                pex_repository=resolver_configuration.pex_repository
            )
        ):
            return resolve_from_pex(
                targets=targets,
                pex=resolver_configuration.pex_repository,
                requirements=requirement_configuration.requirements,
                requirement_files=requirement_configuration.requirement_files,
                constraint_files=requirement_configuration.constraint_files,
                network_configuration=resolver_configuration.network_configuration,
                transitive=resolver_configuration.transitive,
                ignore_errors=ignore_errors,
            )
    else:
        with TRACER.timed("Resolving requirements."):
            return resolve_via_pip(
                targets=targets,
                requirements=requirement_configuration.requirements,
                requirement_files=requirement_configuration.requirement_files,
                constraint_files=requirement_configuration.constraint_files,
                allow_prereleases=resolver_configuration.allow_prereleases,
                transitive=resolver_configuration.transitive,
                indexes=resolver_configuration.repos_configuration.indexes,
                find_links=resolver_configuration.repos_configuration.find_links,
                resolver_version=resolver_configuration.resolver_version,
                network_configuration=resolver_configuration.network_configuration,
                password_entries=resolver_configuration.repos_configuration.password_entries,
                build=resolver_configuration.allow_builds,
                use_wheel=resolver_configuration.allow_wheels,
                prefer_older_binary=resolver_configuration.prefer_older_binary,
                use_pep517=resolver_configuration.use_pep517,
                build_isolation=resolver_configuration.build_isolation,
                compile=compile_pyc,
                max_parallel_jobs=resolver_configuration.max_jobs,
                ignore_errors=ignore_errors,
                preserve_log=resolver_configuration.preserve_log,
                pip_version=resolver_configuration.version,
                resolver=ConfiguredResolver(pip_configuration=resolver_configuration),
            )



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import shlex
import subprocess
import sys
from distutils import log
from distutils.core import Command
from io import StringIO

from pex.bin.pex import configure_clp
from pex.common import die
from pex.compatibility import ConfigParser, string, to_unicode
from pex.interpreter import PythonInterpreter


# Suppress checkstyle violations due to distutils command requirements.
class bdist_pex(Command):  # noqa
    @staticmethod
    def get_log_level():
        # A hack to get the existing distutils logging level.
        existing_level = log.set_threshold(log.INFO)
        log.set_threshold(existing_level)

        return existing_level

    description = "create a PEX file from a source distribution"  # noqa

    user_options = [  # noqa
        ("bdist-all", None, "pexify all defined entry points"),
        ("bdist-dir=", None, "the directory into which pexes will be written, default: dist."),
        ("pex-args=", None, "additional arguments to the pex tool"),
    ]

    boolean_options = [  # noqa
        "bdist-all",
    ]

    def initialize_options(self):
        self.bdist_all = False
        self.bdist_dir = None
        self.pex_args = ""

    def finalize_options(self):
        self.pex_args = shlex.split(self.pex_args)

    def parse_entry_points(self):
        def parse_entry_point_name(entry_point):
            script_name = entry_point.split("=", 1)[0]
            return script_name.strip()

        raw_entry_points = self.distribution.entry_points

        if isinstance(raw_entry_points, string):
            parser = ConfigParser()
            parser.readfp(StringIO(to_unicode(raw_entry_points)))
            if parser.has_section("console_scripts"):
                return tuple(parser.options("console_scripts"))
        elif isinstance(raw_entry_points, dict):
            try:
                return tuple(
                    parse_entry_point_name(script)
                    for script in raw_entry_points.get("console_scripts", [])
                )
            except ValueError:
                pass
        elif raw_entry_points is not None:
            die("When entry_points is provided, it must be a string or dict.")

        return ()

    def run(self):
        parser = configure_clp()
        options = parser.parse_args(self.pex_args)

        if options.entry_point or options.script or options.pex_name:
            die(
                "Must not specify entry point, script or output file to --pex-args, given: {}".format(
                    " ".join(self.pex_args)
                )
            )

        name = self.distribution.get_name()
        version = self.distribution.get_version()

        package_dir = os.path.dirname(
            os.path.realpath(os.path.expanduser(self.distribution.script_name))
        )
        if self.bdist_dir is None:
            self.bdist_dir = os.path.join(package_dir, "dist")

        console_scripts = self.parse_entry_points()

        pex_specs = []
        if self.bdist_all:
            # Write all entry points into unversioned pex files.
            pex_specs.extend(
                (script_name, os.path.join(self.bdist_dir, script_name))
                for script_name in console_scripts
            )
        else:
            target = os.path.join(self.bdist_dir, name + "-" + version + ".pex")
            pex_specs.append((name if name in console_scripts else None, target))

        args = ["-m", "pex", package_dir] + options.requirements + self.pex_args
        if self.get_log_level() < log.INFO and options.verbosity == 0:
            args.append("-v")

        for script_name, target in pex_specs:
            pex_cmd = args + ["--output-file", target]
            if script_name:
                log.info("Writing %s to %s" % (script_name, target))
                pex_cmd += ["--script", script_name]
            else:
                # The package has no namesake entry point, so build an environment pex.
                log.info("Writing environment pex into %s" % target)

            cmd, process = PythonInterpreter.get().open_process(
                args=pex_cmd,
                stderr=subprocess.PIPE,
                # In order for code to run to here, pex is on the sys.path - make sure to propagate the
                # sys.path so the subprocess can find us.
                pythonpath=sys.path,
            )
            _, stderr = process.communicate()
            result = process.returncode
            if result != 0:
                die(
                    "Failed to create pex via {}:\n{}".format(
                        " ".join(cmd), stderr.decode("utf-8")
                    ),
                    result,
                )
            else:
                log.debug(
                    "Successfully created pex via {}:\n{}".format(
                        " ".join(cmd), stderr.decode("utf-8")
                    )
                )



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List


@pytest.mark.parametrize(
    ["boot_args"],
    [
        pytest.param([], id="__main__.py boot"),
        pytest.param(["--sh-boot"], id="--sh-boot"),
    ],
)
def test_venv_pex_script_non_hermetic(
    tmpdir,  # type: Any
    boot_args,  # type: List[str]
):
    # type: (...) -> None

    # A console script that injects an element in the PYTHONPATH.
    ot_simulator_src = os.path.join(str(tmpdir), "src")
    with safe_open(os.path.join(ot_simulator_src, "ot_simulator.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import os
                import sys

                def run():
                    pythonpath = ["injected"]
                    existing_pythonpath = os.environ.get("PYTHONPATH")
                    if existing_pythonpath:
                        pythonpath.extend(existing_pythonpath.split(os.pathsep))
                    os.environ["PYTHONPATH"] = os.pathsep.join(pythonpath)

                    os.execv(sys.argv[1], sys.argv[1:])
                """
            )
        )
    with safe_open(os.path.join(ot_simulator_src, "setup.cfg"), "w") as fp:
        fp.write(
            dedent(
                """\
                [metadata]
                name = ot-simulator
                version = 0.0.1

                [options]
                py_modules =
                    ot_simulator

                [options.entry_points]
                console_scripts =
                    instrument = ot_simulator:run
                """
            )
        )
    with safe_open(os.path.join(ot_simulator_src, "setup.py"), "w") as fp:
        fp.write("from setuptools import setup; setup()")

    # An entrypoint that can observe the PYTHONPATH / sys.path.
    app = os.path.join(str(tmpdir), "app.exe")
    with safe_open(app, "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import os
                import sys

                json.dump(
                    {
                        "PYTHONPATH": os.environ.get("PYTHONPATH"),
                        "sys.path": sys.path
                    },
                    sys.stdout
                )
                """
            )
        )

    pex_root = os.path.join(str(tmpdir), "pex_root")

    def create_app_pex(hermetic_scripts):
        # type: (bool) -> str
        pex = os.path.join(
            str(tmpdir), "{}-app.pex".format("hermetic" if hermetic_scripts else "non-hermetic")
        )
        argv = [
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            ot_simulator_src,
            "--exe",
            app,
            "--venv",
            "-o",
            pex,
        ] + boot_args
        if not hermetic_scripts:
            argv.append("--non-hermetic-venv-scripts")
        run_pex_command(argv).assert_success()
        return pex

    cwd = os.path.join(str(tmpdir), "cwd")
    os.mkdir(cwd)

    # A standard hermetic venv pex should be able to see PYTHONPATH but not have its sys.path
    # tainted by it.
    hermetic_app_pex = create_app_pex(hermetic_scripts=True)
    hermetic = json.loads(
        subprocess.check_output(
            args=[hermetic_app_pex], cwd=cwd, env=make_env(PYTHONPATH="ambient")
        )
    )
    assert "ambient" == hermetic["PYTHONPATH"]
    assert os.path.join(cwd, "ambient") not in hermetic["sys.path"]

    # A non-hermetic venv pex should be able to both see PYTHONPATH and have it affect its sys.path.
    non_hermetic_app_pex = create_app_pex(hermetic_scripts=False)
    baseline = json.loads(
        subprocess.check_output(
            args=[non_hermetic_app_pex], cwd=cwd, env=make_env(PYTHONPATH="ambient")
        )
    )
    assert "ambient" == baseline["PYTHONPATH"]
    assert os.path.join(cwd, "ambient") in baseline["sys.path"]

    # A non-hermetic venv pex should have the non-hermeticity extend to its console scripts in
    # addition to the main entry point `pex` script.
    instrumented = json.loads(
        subprocess.check_output(
            args=[non_hermetic_app_pex, non_hermetic_app_pex],
            cwd=cwd,
            env=make_env(PYTHONPATH="ambient", PEX_SCRIPT="instrument"),
        )
    )
    assert "injected:ambient" == instrumented["PYTHONPATH"]
    assert sorted(baseline["sys.path"] + [os.path.join(cwd, "injected")]) == sorted(
        instrumented["sys.path"]
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import errno
import os
import re
from threading import Condition, Event, Thread
from types import TracebackType

from pex.compatibility import get_stdout_bytes_buffer
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, BinaryIO, Iterable, Optional, Pattern, Type


class Tailer(Thread):
    @classmethod
    def tail(
        cls,
        path,  # type: str
        encoding="utf-8",  # type: str
        filters=(),  # type: Iterable[Pattern]
        poll_interval=0.1,  # type: float
        output=get_stdout_bytes_buffer(),  # type: BinaryIO
    ):
        # type: (...) -> Tailer
        tailer = cls(
            path, encoding=encoding, filters=filters, poll_interval=poll_interval, output=output
        )
        tailer.start()
        return tailer

    def __init__(
        self,
        path,  # type: str
        encoding="utf-8",  # type: str
        filters=(),  # type: Iterable[Pattern]
        poll_interval=0.1,  # type: float
        output=get_stdout_bytes_buffer(),  # type: BinaryIO
    ):
        # type: (...) -> None
        super(Tailer, self).__init__(name="Tailing {path}".format(path=path))
        self.daemon = True

        self._path = path
        self._encoding = encoding
        self._filters = filters or [re.compile(r".*")]
        self._poll_interval = poll_interval
        self._output = output

        self._offset = 0
        self._observer = Condition()
        self._stopped = Event()

    def __enter__(self):
        # type: () -> Tailer
        if not self._stopped.is_set() and not self.is_alive():
            self.start()
        return self

    def __exit__(
        self,
        _exc_type,  # type: Optional[Type]
        _exc_val,  # type: Optional[Any]
        _exc_tb,  # type: Optional[TracebackType]
    ):
        # type: (...) -> None
        self.stop()

    def run(self):
        # type: () -> None
        while not self._stopped.is_set():
            if not self._tail():
                self._stopped.wait(self._poll_interval)
        self._tail()

    def _tail(self):
        # type: () -> bool
        try:
            offset = os.path.getsize(self._path)
            if offset <= self._offset:
                # We do not handle generic tailing here where the file may be truncated and then
                # re-grow. The Pip log case is a simple always growing file.
                return False
        except OSError as e:
            if e.errno == errno.ENOENT:
                # The file may not exist yet; so we just wait for it to appear.
                return False
            raise e

        with open(self._path, "rb") as fp:
            fp.seek(self._offset)
            while fp.tell() < offset:
                line_bytes = fp.readline()
                line = line_bytes.decode(self._encoding)
                for pattern in self._filters:
                    match = pattern.match(line)
                    if match:
                        if match.groups():
                            eol = re.search(r"(?P<eol>\r\n|\r|\n)$", line)
                            self._output.write(
                                "{groups}{eol}".format(
                                    groups="".join(match.groups()),
                                    eol=eol.group("eol") if eol else "",
                                ).encode(self._encoding)
                            )
                        else:
                            self._output.write(line_bytes)
                        break
            self._offset = fp.tell()

        self._notify_observers()
        return True

    def _notify_observers(self):
        # type: () -> None
        with self._observer:
            self._observer.notify_all()

    def observe(self, timeout=None):
        # type: (Optional[float]) -> None
        """Wait for at least one new line of tailer output to be observable.

        Waits forever unless `timeout` is specified, in which case wait at most that many seconds,
        but returns immediately if the tailer is stopped.
        """
        if not self._stopped.is_set():
            with self._observer:
                self._observer.wait(timeout=timeout)

    def stop(self):
        # type: () -> None
        self._stopped.set()
        self.join()



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.interpreter import PythonInterpreter
from pex.pep_376 import InstalledFile, find_and_replace_path_components
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def test_filter_path_invalid():
    # type: () -> None

    with pytest.raises(ValueError):
        find_and_replace_path_components("foo", "bar", "")
        find_and_replace_path_components("foo", "", "baz")


def test_filter_path_noop():
    # type: () -> None

    assert "" == find_and_replace_path_components("", "spam", "eggs")
    assert "." == find_and_replace_path_components(".", "spam", "eggs")
    assert ".." == find_and_replace_path_components("..", "spam", "eggs")
    assert "/" == find_and_replace_path_components("/", "spam", "eggs")
    assert "foo/bar/baz" == find_and_replace_path_components("foo/bar/baz", "spam", "eggs")


def test_filter_path_basic():
    # type: () -> None

    assert "spam/bar/baz" == find_and_replace_path_components("foo/bar/baz", "foo", "spam")
    assert "foo/spam/baz" == find_and_replace_path_components("foo/bar/baz", "bar", "spam")
    assert "foo/bar/spam" == find_and_replace_path_components("foo/bar/baz", "baz", "spam")


def test_filter_path_absolute():
    # type: () -> None

    assert "/spam/bar/baz" == find_and_replace_path_components("/foo/bar/baz", "foo", "spam")


def test_filter_path_relative():
    # type: () -> None

    assert "../spam/bar/baz" == find_and_replace_path_components("../foo/bar/baz", "foo", "spam")
    assert "./spam/bar/baz" == find_and_replace_path_components("./foo/bar/baz", "foo", "spam")
    assert "/spam/../bar/./baz" == find_and_replace_path_components(
        "/foo/../bar/./baz", "foo", "spam"
    )


def test_installed_file_path_normalization_noop(
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    def assert_noop(interpreter=None):
        # type: (Optional[PythonInterpreter]) -> None
        assert "foo/bar" == InstalledFile.normalized_path("foo/bar", interpreter=interpreter)
        assert "foo/python2.0" == InstalledFile.normalized_path(
            "foo/python2.0", interpreter=interpreter
        )
        assert "foo/bar" == InstalledFile.denormalized_path("foo/bar", interpreter=interpreter)

    assert_noop()
    assert_noop(py38)
    assert_noop(py310)


def test_installed_file_path_normalization_nominal(
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
):
    # type: (...) -> None

    assert "foo/pythonX.Y/bar" == InstalledFile.normalized_path(
        "foo/python3.8/bar", interpreter=py38
    )
    assert "foo/pythonX.Y/bar" == InstalledFile.normalized_path(
        "foo/python3.10/bar", interpreter=py310
    )

    assert "foo/python3.8/bar" == InstalledFile.denormalized_path(
        "foo/pythonX.Y/bar", interpreter=py38
    )
    assert "foo/python3.10/bar" == InstalledFile.denormalized_path(
        "foo/pythonX.Y/bar", interpreter=py310
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
import sys

import pytest

from pex.interpreter import PythonInterpreter
from pex.pex import PEX
from pex.testing import PY27, ensure_python_interpreter, run_pex_command
from pex.third_party.packaging import tags
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.fixture
def supported_py27():
    # type: (...) -> str

    python = ensure_python_interpreter(PY27)

    # JPype1 version 0.7.0 has CPython 2.7 published wheels with tags:
    # + cp27-cp27m-manylinux2010_x86_64
    # + cp27-cp27mu-manylinux2010_x86_64
    python_identity = PythonInterpreter.from_binary(python).identity
    tag = "{python}-{abi}-manylinux2010_x86_64".format(
        python=python_identity.python_tag,
        abi=python_identity.abi_tag,
    )
    if len(python_identity.supported_tags.compatible_tags(tags.parse_tag(tag))) == 0:
        pytest.skip("Test requires a manylinux2010 x86_64 compatible platform")
    return python


def test_prefer_binary(
    tmpdir,  # type: Any
    supported_py27,  # type: str
):
    # type: (...) -> None

    result = run_pex_command(args=["JPype1"], python=supported_py27)
    result.assert_failure()
    assert "ImportError: No module named pathlib" in result.error, (
        "The latest versions of JPype1 do not support Python 2.7 and their setup.py use Python 3 "
        "only features; so we expect an unconstrained resolve to pick the latest JPype1 sdist,"
        "attempt to build a wheel from it and fail."
    )

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=["--prefer-binary", "JPype1", "-o", pex], python=supported_py27
    ).assert_success()

    subprocess.check_call(args=[supported_py27, pex, "-c", "import jpype"])
    distributions = tuple(
        PEX(pex, interpreter=PythonInterpreter.from_binary(supported_py27)).resolve()
    )
    assert 1 == len(distributions)
    assert "JPype1" == distributions[0].project_name
    assert (
        "0.7.0" == distributions[0].version
    ), "The last version of JPype1 to publish a Python 2.7 compatible distribution was 0.7.0"



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import re

import pytest

from pex.pep_425 import CompatibilityTags
from pex.third_party.packaging import tags
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterable


def equatable(tags):
    # type: (Iterable[tags.Tag]) -> Any
    return sorted(map(str, tags))


def test_tags_from_wheel_nominal():
    # type: () -> None

    assert equatable(tags.parse_tag("py2.py3-none-any")) == equatable(
        CompatibilityTags.from_wheel("foo-1.2.3-py2.py3-none-any.whl")
    )
    assert equatable(CompatibilityTags.from_strings(["py2-none-any", "py3-none-any"])) == equatable(
        CompatibilityTags.from_wheel("foo-1.2.3-py2.py3-none-any.whl")
    )
    assert equatable(tags.parse_tag("py2.py3-none-any")) == equatable(
        CompatibilityTags.from_wheel("foo-1.2.3-build_tag-py2.py3-none-any.whl")
    )
    assert equatable(tags.parse_tag("py3-none-any")) == equatable(
        CompatibilityTags.from_wheel("path/to/bar-4.5.6-py3-none-any.whl")
    )


def test_tags_from_wheel_invalid():
    # type: () -> None

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Can only calculate wheel tags from a filename that ends in .whl per "
            "https://peps.python.org/pep-0427/#file-name-convention, given: "
            "'foo-1.2.3-py2.py3-none-any'"
        ),
    ):
        CompatibilityTags.from_wheel("foo-1.2.3-py2.py3-none-any")

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Can only calculate wheel tags from a filename that ends in "
            "`-{python tag}-{abi tag}-{platform tag}.whl` per "
            "https://peps.python.org/pep-0427/#file-name-convention, given: 'py2.py3-none-any.whl'"
        ),
    ):
        CompatibilityTags.from_wheel("py2.py3-none-any.whl")



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.typing import TYPE_CHECKING, Generic, cast, overload

if TYPE_CHECKING:
    from typing import Any, Iterable, Iterator, Optional, Protocol, TypeVar, Union

    class Comparable(Protocol):
        def __lt__(self, other):
            # type: (Any) -> bool
            pass

    _CT = TypeVar("_CT", bound=Comparable)

    _T = TypeVar("_T", bound=Comparable)

    class Comparator(Protocol):
        def __call__(self, item):
            # type: (Any) -> Comparable
            pass


class SortedTuple(Generic["_CT"], tuple):
    @overload
    def __new__(cls):
        # type: () -> SortedTuple[Any]
        pass

    @overload
    def __new__(
        cls,
        iterable,  # type: Iterable[_CT]
        key=None,  # type: None
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_CT]
        pass

    @overload
    def __new__(
        cls,
        iterable,  # type: Iterable[Any]
        key,  # type: Comparator
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_CT]
        pass

    def __new__(
        cls,
        iterable=None,  # type: Union[None, Iterable[_CT], Iterable[Any]]
        key=None,  # type: Optional[Comparator]
        reverse=False,  # type: bool
    ):
        # type: (...) -> SortedTuple[_CT]
        return super(SortedTuple, cls).__new__(
            cls, sorted(iterable, key=key, reverse=reverse) if iterable else ()
        )

    @overload
    def __getitem__(self, index):
        # type: (int) -> _CT
        pass

    @overload
    def __getitem__(self, slice_spec):
        # type: (slice) -> SortedTuple[_CT]
        pass

    def __getitem__(self, item):
        # type: (Union[int, slice]) -> Union[_CT, SortedTuple[_CT]]
        return cast("Union[_CT, SortedTuple[_CT]]", tuple.__getitem__(self, item))

    def __iter__(self):
        # type: () -> Iterator[_CT]
        return tuple.__iter__(self)



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.pip.installation import compatible_version, validate_targets
from pex.pip.version import PipVersionValue
from pex.resolve.resolver_configuration import (
    LockRepositoryConfiguration,
    PexRepositoryConfiguration,
    PipConfiguration,
)
from pex.result import Error, catch, try_
from pex.targets import Targets
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Optional, TypeVar, Union

    import attr  # vendor:skip

    Configuration = Union[LockRepositoryConfiguration, PexRepositoryConfiguration, PipConfiguration]
    _C = TypeVar("_C", bound=Configuration)

else:
    from pex.third_party import attr


def _finalize_pip_configuration(
    pip_configuration,  # type: PipConfiguration
    targets,  # type: Targets
    context,  # type: str
    pip_version=None,  # type: Optional[PipVersionValue]
):
    # type: (...) -> Union[PipConfiguration, Error]
    version = pip_version or pip_configuration.version
    if pip_configuration.allow_version_fallback:
        return attr.evolve(pip_configuration, version=compatible_version(targets, version, context))

    result = catch(validate_targets, targets, version, context)
    if isinstance(result, Error):
        return result
    return attr.evolve(pip_configuration, version=version)


def finalize(
    resolver_configuration,  # type: _C
    targets,  # type: Targets
    context,  # type: str
    pip_version=None,  # type: Optional[PipVersionValue]
):
    # type: (...) -> Union[_C, Error]

    if isinstance(resolver_configuration, PipConfiguration):
        return cast(
            "_C",
            _finalize_pip_configuration(
                resolver_configuration, targets, context, pip_version=pip_version
            ),
        )

    if isinstance(resolver_configuration, LockRepositoryConfiguration):
        lock_file = try_(resolver_configuration.parse_lock())
        pip_configuration = try_(
            _finalize_pip_configuration(
                resolver_configuration.pip_configuration,
                targets,
                context,
                pip_version=pip_version or lock_file.pip_version,
            )
        )
        return cast(
            "_C",
            attr.evolve(
                resolver_configuration,
                parse_lock=lambda: lock_file,
                pip_configuration=pip_configuration,
            ),
        )

    return resolver_configuration



"""
Compatibility Support for Python 2.7 and earlier
"""

import sys
import platform

if "__PEX_UNVENDORED__" in __import__("os").environ:
  from setuptools.extern import six  # vendor:skip
else:
  from pex.third_party.setuptools.extern import six



def get_all_headers(message, key):
    """
    Given an HTTPMessage, return all headers matching a given key.
    """
    return message.get_all(key)


if six.PY2:
    def get_all_headers(message, key):
        return message.getheaders(key)


linux_py2_ascii = (
    platform.system() == 'Linux' and
    six.PY2
)

rmtree_safe = str if linux_py2_ascii else lambda x: x
"""Workaround for http://bugs.python.org/issue24672"""


try:
    from ._imp import find_module, PY_COMPILED, PY_FROZEN, PY_SOURCE
    from ._imp import get_frozen_object, get_module
except ImportError:
    import imp
    from imp import PY_COMPILED, PY_FROZEN, PY_SOURCE  # noqa

    def find_module(module, paths=None):
        """Just like 'imp.find_module()', but with package support"""
        parts = module.split('.')
        while parts:
            part = parts.pop(0)
            f, path, (suffix, mode, kind) = info = imp.find_module(part, paths)

            if kind == imp.PKG_DIRECTORY:
                parts = parts or ['__init__']
                paths = [path]

            elif parts:
                raise ImportError("Can't find %r in %s" % (parts, module))

        return info

    def get_frozen_object(module, paths):
        return imp.get_frozen_object(module)

    def get_module(module, paths, info):
        imp.load_module(module, *info)
        return sys.modules[module]



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import base64
import csv
import errno
import hashlib
import itertools
import json
import os
import shutil
from contextlib import closing
from fileinput import FileInput

from pex import dist_metadata, hashing
from pex.common import filter_pyc_dirs, filter_pyc_files, is_python_script, safe_mkdir, safe_open
from pex.compatibility import get_stdout_bytes_buffer, urlparse
from pex.dist_metadata import Distribution, EntryPoint
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING, cast
from pex.venv.virtualenv import Virtualenv

if TYPE_CHECKING:
    from typing import (
        Callable,
        Container,
        Dict,
        Iterable,
        Iterator,
        List,
        Optional,
        Protocol,
        Tuple,
        Union,
    )

    import attr  # vendor:skip

    from pex.hashing import Hasher

    class CSVWriter(Protocol):
        def writerow(self, row):
            # type: (Iterable[Union[str, int]]) -> None
            pass

else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Digest(object):
    algorithm = attr.ib()  # type: str
    encoded_hash = attr.ib()  # type: str

    def new_hasher(self):
        # type: () -> Hasher
        return hashlib.new(self.algorithm)


@attr.s(frozen=True)
class Hash(object):
    @classmethod
    def create(cls, hasher):
        # type: (Hasher) -> Hash

        # The fingerprint encoding is defined for PEP-376 RECORD files as `urlsafe-base64-nopad`
        # which is fully spelled out in code in PEP-427:
        # + https://peps.python.org/pep-0376/#record
        # + https://peps.python.org/pep-0427/#appendix
        fingerprint = base64.urlsafe_b64encode(hasher.digest()).rstrip(b"=")
        return cls(value="{alg}={hash}".format(alg=hasher.name, hash=fingerprint.decode("ascii")))

    value = attr.ib()  # type: str

    def __str__(self):
        # type: () -> str
        return self.value


def find_and_replace_path_components(
    path,  # type: str
    find,  # type: str
    replace,  # type: str
):
    # type: (...) -> str
    """Replace components of `path` that are exactly `find` with `replace`.

    >>> find_and_replace_path_components("foo/bar/baz", "bar", "spam")
    foo/spam/baz
    >>>
    """
    if not find or not replace:
        raise ValueError(
            "Both find and replace must be non-empty strings. Given find={find!r} "
            "replace={replace!r}".format(find=find, replace=replace)
        )
    if not path:
        return path

    components = []
    head = path
    while head:
        new_head, tail = os.path.split(head)
        if new_head == head:
            components.append(head)
            break
        components.append(tail)
        head = new_head
    components.reverse()
    return os.path.join(*(replace if component == find else component for component in components))


@attr.s(frozen=True)
class InstalledFile(object):
    """The record of a single installed file from a PEP 376 RECORD file.

    See: https://www.python.org/dev/peps/pep-0376/#record
    """

    _PYTHON_VER_PLACEHOLDER = "pythonX.Y"

    @staticmethod
    def _python_ver(interpreter=None):
        # type: (Optional[PythonInterpreter]) -> str
        python = interpreter or PythonInterpreter.get()
        return "python{major}.{minor}".format(major=python.version[0], minor=python.version[1])

    @classmethod
    def normalized_path(
        cls,
        path,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> str
        return find_and_replace_path_components(
            path, cls._python_ver(interpreter=interpreter), cls._PYTHON_VER_PLACEHOLDER
        )

    @classmethod
    def denormalized_path(
        cls,
        path,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> str
        return find_and_replace_path_components(
            path, cls._PYTHON_VER_PLACEHOLDER, cls._python_ver(interpreter=interpreter)
        )

    path = attr.ib()  # type: str
    hash = attr.ib(default=None)  # type: Optional[Hash]
    size = attr.ib(default=None)  # type: Optional[int]


class InstalledWheelError(Exception):
    pass


class LoadError(InstalledWheelError):
    """Indicates an installed wheel was not loadable at a particular path."""


class ReinstallError(InstalledWheelError):
    """Indicates an error re-installing an installed wheel."""


@attr.s(frozen=True)
class InstalledWheel(object):
    _LAYOUT_JSON_FILENAME = ".layout.json"

    @classmethod
    def layout_file(cls, prefix_dir):
        # type: (str) -> str
        return os.path.join(prefix_dir, cls._LAYOUT_JSON_FILENAME)

    @classmethod
    def save(
        cls,
        prefix_dir,  # type: str
        stash_dir,  # type: str
        record_relpath,  # type: str
    ):
        # type: (...) -> InstalledWheel
        layout = {"stash_dir": stash_dir, "record_relpath": record_relpath}
        with open(cls.layout_file(prefix_dir), "w") as fp:
            json.dump(layout, fp, sort_keys=True)
        return cls(prefix_dir=prefix_dir, stash_dir=stash_dir, record_relpath=record_relpath)

    @classmethod
    def load(cls, prefix_dir):
        # type: (str) -> InstalledWheel
        layout_file = cls.layout_file(prefix_dir)
        try:
            with open(layout_file) as fp:
                layout = json.load(fp)
        except (IOError, OSError) as e:
            raise LoadError(
                "Failed to load an installed wheel layout from {layout_file}: {err}".format(
                    layout_file=layout_file, err=e
                )
            )
        if not isinstance(layout, dict):
            raise LoadError(
                "The installed wheel layout file at {layout_file} must contain a single top-level "
                "object, found: {value}.".format(layout_file=layout_file, value=layout)
            )
        stash_dir = layout.get("stash_dir")
        record_relpath = layout.get("record_relpath")
        if not stash_dir or not record_relpath:
            raise LoadError(
                "The installed wheel layout file at {layout_file} must contain an object with both "
                "`stash_dir` and `record_relpath` attributes, found: {value}".format(
                    layout_file=layout_file, value=layout
                )
            )
        return cls(
            prefix_dir=prefix_dir,
            stash_dir=cast(str, stash_dir),
            record_relpath=cast(str, record_relpath),
        )

    prefix_dir = attr.ib()  # type: str
    stash_dir = attr.ib()  # type: str
    record_relpath = attr.ib()  # type: str

    def stashed_path(self, *components):
        # type: (*str) -> str
        return os.path.join(self.prefix_dir, self.stash_dir, *components)

    @staticmethod
    def _create_installed_file(
        path,  # type: str
        dest_dir,  # type: str
    ):
        # type: (...) -> InstalledFile
        hasher = hashlib.sha256()
        hashing.file_hash(path, digest=hasher)
        return InstalledFile(
            path=os.path.relpath(path, dest_dir),
            hash=Hash.create(hasher),
            size=os.stat(path).st_size,
        )

    def create_record(
        self,
        dst,  # type: str
        installed_files,  # type: Iterable[InstalledFile]
    ):
        # type: (...) -> None

        # The RECORD is a csv file with the path to each installed file in the 1st column.
        # See: https://www.python.org/dev/peps/pep-0376/#record
        with safe_open(os.path.join(dst, self.record_relpath), "w") as fp:
            csv_writer = cast(
                "CSVWriter",
                csv.writer(fp, delimiter=",", quotechar='"', lineterminator="\n"),
            )
            for installed_file in sorted(installed_files, key=lambda installed: installed.path):
                csv_writer.writerow(attr.astuple(installed_file, recurse=False))

    def reinstall_flat(
        self,
        target_dir,  # type: str
        symlink=False,  # type: bool
    ):
        # type: (...) -> Iterator[Tuple[str, str]]
        """Re-installs the installed wheel in a flat target directory.

        N.B.: A record of reinstalled files is returned in the form of an iterator that must be
        consumed to drive the installation to completion.

        If there is an error re-installing a file due to it already existing in the target
        directory, the error is suppressed, and it's expected that the caller detects this by
        comparing the record of installed files against those installed previously.

        :return: An iterator over src -> dst pairs.
        """
        installed_files = [InstalledFile(self.record_relpath)]
        for src, dst in itertools.chain(
            self._reinstall_stash(dest_dir=target_dir),
            self._reinstall_site_packages(target_dir, symlink=symlink),
        ):
            installed_files.append(self._create_installed_file(path=dst, dest_dir=target_dir))
            yield src, dst

        self.create_record(target_dir, installed_files)

    def reinstall_venv(
        self,
        venv,  # type: Virtualenv
        symlink=False,  # type: bool
        rel_extra_path=None,  # type: Optional[str]
    ):
        # type: (...) -> Iterator[Tuple[str, str]]
        """Re-installs the installed wheel in a venv.

        N.B.: A record of reinstalled files is returned in the form of an iterator that must be
        consumed to drive the installation to completion.

        If there is an error re-installing a file due to it already existing in the destination
        venv, the error is suppressed, and it's expected that the caller detects this by comparing
        the record of installed files against those installed previously.

        :return: An iterator over src -> dst pairs.
        """
        site_packages_dir = (
            os.path.join(venv.site_packages_dir, rel_extra_path)
            if rel_extra_path
            else venv.site_packages_dir
        )

        installed_files = [InstalledFile(self.record_relpath)]
        for src, dst in itertools.chain(
            self._reinstall_stash(dest_dir=venv.venv_dir, interpreter=venv.interpreter),
            self._reinstall_site_packages(site_packages_dir, symlink=symlink),
        ):
            installed_files.append(
                self._create_installed_file(path=dst, dest_dir=site_packages_dir)
            )
            yield src, dst

        self.create_record(site_packages_dir, installed_files)

    def _reinstall_stash(
        self,
        dest_dir,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> Iterator[Tuple[str, str]]

        link = True
        stash_abs_path = os.path.join(self.prefix_dir, self.stash_dir)
        for root, dirs, files in os.walk(stash_abs_path, topdown=True, followlinks=True):
            dir_created = False
            for f in files:
                src = os.path.join(root, f)
                src_relpath = os.path.relpath(src, stash_abs_path)
                dst = InstalledFile.denormalized_path(
                    path=os.path.join(dest_dir, src_relpath), interpreter=interpreter
                )
                if not dir_created:
                    safe_mkdir(os.path.dirname(dst))
                    dir_created = True
                try:
                    # We only try to link regular files since linking a symlink on Linux can produce
                    # another symlink, which leaves open the possibility the src target could later
                    # go missing leaving the dst dangling.
                    if link and not os.path.islink(src):
                        try:
                            os.link(src, dst)
                            continue
                        except OSError as e:
                            if e.errno != errno.EXDEV:
                                raise e
                            link = False
                    shutil.copy(src, dst)
                except (IOError, OSError) as e:
                    if e.errno != errno.EEXIST:
                        raise e
                finally:
                    yield src, dst

    def _reinstall_site_packages(
        self,
        site_packages_dir,  # type: str
        symlink=False,  # type: bool
    ):
        # type: (...) -> Iterator[Tuple[str, str]]

        link = True
        for root, dirs, files in os.walk(self.prefix_dir, topdown=True, followlinks=True):
            if root == self.prefix_dir:
                dirs[:] = [d for d in filter_pyc_dirs(dirs) if d != self.stash_dir]
                files[:] = [f for f in filter_pyc_files(files) if f != self._LAYOUT_JSON_FILENAME]

            traverse = set(dirs)
            for path, is_dir in itertools.chain(
                zip(dirs, itertools.repeat(True)), zip(files, itertools.repeat(False))
            ):
                src_entry = os.path.join(root, path)
                dst_entry = os.path.join(
                    site_packages_dir, os.path.relpath(src_entry, self.prefix_dir)
                )
                try:
                    if symlink and not (
                        src_entry.endswith(".dist-info") and os.path.isdir(src_entry)
                    ):
                        dst_parent = os.path.dirname(dst_entry)
                        safe_mkdir(dst_parent)
                        rel_src = os.path.relpath(src_entry, dst_parent)
                        os.symlink(rel_src, dst_entry)
                        traverse.discard(path)
                    elif is_dir:
                        safe_mkdir(dst_entry)
                    else:
                        # We only try to link regular files since linking a symlink on Linux can
                        # produce another symlink, which leaves open the possibility the src_entry
                        # target could later go missing leaving the dst_entry dangling.
                        if link and not os.path.islink(src_entry):
                            try:
                                os.link(src_entry, dst_entry)
                                continue
                            except OSError as e:
                                if e.errno != errno.EXDEV:
                                    raise e
                                link = False
                        shutil.copy(src_entry, dst_entry)
                except (IOError, OSError) as e:
                    if e.errno != errno.EEXIST:
                        raise e
                finally:
                    if not is_dir:
                        yield src_entry, dst_entry

            dirs[:] = list(traverse)


class RecordError(Exception):
    pass


class RecordNotFoundError(RecordError):
    """Indicates a distribution's RECORD metadata could not be found."""


class UnrecognizedInstallationSchemeError(RecordError):
    """Indicates a distribution's RECORD was nested in an unrecognized installation scheme."""


@attr.s(frozen=True)
class Record(object):
    """Represents the PEP-376 RECORD of an installed wheel.

    See: https://www.python.org/dev/peps/pep-0376/#record
    """

    @classmethod
    def read(
        cls,
        lines,  # type: Union[FileInput[str], Iterator[str]]
        exclude=None,  # type: Optional[Callable[[str], bool]]
    ):
        # type: (...) -> Iterator[InstalledFile]

        # The RECORD is a csv file with the path to each installed file in the 1st column.
        # See: https://www.python.org/dev/peps/pep-0376/#record
        for line, (path, fingerprint, file_size) in enumerate(
            csv.reader(lines, delimiter=",", quotechar='"'), start=1
        ):
            resolved_path = path
            if exclude and exclude(resolved_path):
                continue
            file_hash = Hash(fingerprint) if fingerprint else None
            size = int(file_size) if file_size else None
            yield InstalledFile(path=path, hash=file_hash, size=size)

    @staticmethod
    def _find_installation(
        prefix_dir,  # type: str
        project_name,  # type: str
        version,  # type: str
    ):
        # type: (...) -> Optional[Tuple[str, str, List[str]]]

        # Some distributions in the wild (namely python-certifi-win32 1.6.1,
        # see: https://github.com/pantsbuild/pex/issues/1861) create their own directories named
        # `site-packages` that are not in-fact located in site-packages (the "purelib" or "platlib"
        # sysconfig install paths). Work around these broken packages by just looking for all
        # `site-packages` subdirectories of the `prefix_dir` and checking each for the installation
        # `RECORD`. There should always be just one such installation `RECORD` resulting from a
        # `pip install --prefix <prefix_dir> --no-deps <wheel file>` and so this is safe.
        site_packages_dirs = [
            os.path.join(root, d)
            for root, dirs, _ in os.walk(prefix_dir)
            for d in dirs
            if d == "site-packages"
        ]
        for site_packages_dir in site_packages_dirs:
            site_packages_listing = [
                os.path.relpath(os.path.join(root, f), site_packages_dir)
                for root, _, files in os.walk(site_packages_dir)
                for f in files
            ]
            record_relative_path = dist_metadata.find_dist_info_file(
                project_name, version=version, filename="RECORD", listing=site_packages_listing
            )
            if record_relative_path:
                return record_relative_path, site_packages_dir, site_packages_listing
        return None

    @classmethod
    def from_prefix_install(
        cls,
        prefix_dir,  # type: str
        project_name,  # type: str
        version,  # type: str
    ):
        result = cls._find_installation(prefix_dir, project_name, version)
        if not result:
            raise RecordNotFoundError(
                "Could not find the installation RECORD for {project_name} {version} under "
                "{prefix_dir}".format(
                    project_name=project_name, version=version, prefix_dir=prefix_dir
                )
            )

        record_relative_path, site_packages, site_packages_listing = result
        metadata_dir = os.path.dirname(record_relative_path)
        base_dir = os.path.relpath(site_packages, prefix_dir)
        return cls(
            project_name=project_name,
            version=version,
            prefix_dir=prefix_dir,
            rel_base_dir=base_dir,
            relative_path=record_relative_path,
            metadata_listing=tuple(
                path for path in site_packages_listing if metadata_dir == os.path.dirname(path)
            ),
        )

    project_name = attr.ib()  # type: str
    version = attr.ib()  # type: str
    prefix_dir = attr.ib()  # type: str
    rel_base_dir = attr.ib()  # type: str
    relative_path = attr.ib()  # type: str
    _metadata_listing = attr.ib()  # type: Tuple[str, ...]

    def _find_dist_info_file(self, filename):
        # type: (str) -> Optional[str]
        metadata_file = dist_metadata.find_dist_info_file(
            project_name=self.project_name,
            version=self.version,
            filename=filename,
            listing=self._metadata_listing,
        )
        if not metadata_file:
            return None
        return os.path.join(self.rel_base_dir, metadata_file)

    def fixup_install(
        self,
        exclude=(),  # type: Container[str]
        interpreter=None,  # type: Optional[PythonInterpreter]
    ):
        # type: (...) -> InstalledWheel
        """Fixes a wheel install to be reproducible and importable.

        After fixed up, this RECORD can be used to re-install the wheel in a venv with `reinstall`.

        :param exclude: Any top-level items to exclude.
        :param interpreter: The interpreter used to perform the wheel install.
        """
        self._fixup_scripts()
        self._fixup_direct_url()

        # The RECORD is unused in PEX zipapp mode and only needed in venv mode. Since it can contain
        # relative path entries that differ between interpreters - notably pypy for Python < 3.8 has
        # a custom scheme - we just delete the file and create it on-demand for venv re-installs.
        os.unlink(os.path.join(self.prefix_dir, self.rel_base_dir, self.relative_path))

        # An example of the installed wheel chroot we're aiming for:
        # .prefix/bin/...                    # scripts
        # .prefix/include/site/pythonX.Y/... # headers
        # .prefix/share/...                  # data files
        # greenlet/...                       # importables
        # greenlet-1.1.2.dist-info/...       # importables
        stash_dir = ".prefix"
        prefix_stash = os.path.join(self.prefix_dir, stash_dir)
        safe_mkdir(prefix_stash)

        # 1. Move everything into the stash.
        for item in os.listdir(self.prefix_dir):
            if stash_dir == item or item in exclude:
                continue
            shutil.move(os.path.join(self.prefix_dir, item), os.path.join(prefix_stash, item))
        # 2. Normalize all `*/{python ver}` paths to `*/pythonX.Y`
        for root, dirs, _ in os.walk(prefix_stash):
            dirs_to_scan = []
            for d in dirs:
                path = os.path.join(root, d)
                normalized_path = InstalledFile.normalized_path(path, interpreter=interpreter)
                if normalized_path != path:
                    shutil.move(path, normalized_path)
                else:
                    dirs_to_scan.append(d)
            dirs[:] = dirs_to_scan

        # 3. Move `site-packages` content back up to the prefix dir chroot so that content is
        # importable when this prefix dir chroot is added to the `sys.path` in PEX zipapp mode.
        importable_stash = InstalledFile.normalized_path(
            os.path.join(prefix_stash, self.rel_base_dir), interpreter=interpreter
        )
        for importable_item in os.listdir(importable_stash):
            shutil.move(
                os.path.join(importable_stash, importable_item),
                os.path.join(self.prefix_dir, importable_item),
            )
        os.rmdir(importable_stash)

        return InstalledWheel.save(
            prefix_dir=self.prefix_dir,
            stash_dir=stash_dir,
            record_relpath=self.relative_path,
        )

    def _fixup_scripts(self):
        # type: (...) -> None
        bin_dir = os.path.join(self.prefix_dir, "bin")
        if not os.path.isdir(bin_dir):
            return

        console_scripts = {}  # type: Dict[str, EntryPoint]
        entry_points_relpath = self._find_dist_info_file("entry_points.txt")
        if entry_points_relpath:
            entry_points_abspath = os.path.join(self.prefix_dir, entry_points_relpath)
            console_scripts.update(
                Distribution.parse_entry_map(entry_points_abspath).get("console_scripts", {})
            )

        scripts = {}  # type: Dict[str, Optional[bytes]]
        for script_name in os.listdir(bin_dir):
            script_path = os.path.join(bin_dir, script_name)
            if is_python_script(script_path):
                scripts[script_path] = None
            elif script_name in console_scripts:
                # When a wheel is installed by Pip and that wheel contains console_scripts, they are
                # normally written with a faux-shebang of:
                # #!python
                #
                # Pex relies on this hermetic shebang and only ever reifies it when creating venvs.
                #
                # If Pip is being run under a Python executable with a path length >127 characters
                # on Linux though, it writes a shebang / header of:
                # #!/bin/sh
                # '''exec' <too long path to Pip venv python> "$0" "$@"'
                # ' '''
                #
                # That header is immediately followed by the expected console_script shim contents:
                # # -*- coding: utf-8 -*-
                # import re
                # import sys
                # from <ep_module> import <ep_func>
                # if __name__ == '__main__':
                #     sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
                #     sys.exit(main())
                #
                # Instead of guessing that 127 characters is the shebang length limit and using
                # Pip's safety-hatch `/bin/sh` trick, we forcibly re-write the header to be just the
                # expected `#!python` shebang. We detect the end of the header with the known 1st
                # line of console_script shim ~code defined in
                # pex/vendor/_vendored/pip/pip/_vendor/distlib/scripts.py on line 41:
                # https://github.com/pantsbuild/pex/blob/196b4cd5b8dd4b4af2586460530e9a777262be7d/pex/vendor/_vendored/pip/pip/_vendor/distlib/scripts.py#L41
                scripts[script_path] = b"# -*- coding: utf-8 -*-"
        if not scripts:
            return

        with closing(FileInput(files=scripts.keys(), inplace=True, mode="rb")) as script_fi:
            first_non_shebang_line = None  # type: Optional[bytes]
            for line in script_fi:
                buffer = get_stdout_bytes_buffer()
                if script_fi.isfirstline():
                    first_non_shebang_line = scripts[script_fi.filename()]
                    # Ensure python shebangs are reproducible. The only place these can be used is
                    # in venv mode PEXes where the `#!python` placeholder shebang will be re-written
                    # to use the venv's python interpreter.
                    buffer.write(b"#!python\n")
                elif (
                    not first_non_shebang_line
                    or cast(bytes, line).strip() == first_non_shebang_line
                ):
                    # N.B.: These lines include the newline already.
                    buffer.write(cast(bytes, line))
                    first_non_shebang_line = None

    def _fixup_direct_url(self):
        # type: () -> None
        direct_url_relpath = self._find_dist_info_file("direct_url.json")
        if direct_url_relpath:
            direct_url_abspath = os.path.join(self.prefix_dir, direct_url_relpath)
            with open(direct_url_abspath) as fp:
                if urlparse.urlparse(json.load(fp)["url"]).scheme == "file":
                    os.unlink(direct_url_abspath)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

from pex.common import safe_copy, temporary_dir
from pex.testing import built_wheel, make_env, make_source_dir, run_pex_command


def test_requirement_setup_py_with_extras():
    # type: () -> None
    with make_source_dir(
        name="project1", version="1.0.0", extras_require={"foo": ["project2"]}
    ) as project1_dir:
        with built_wheel(name="project2", version="2.0.0") as project2_bdist:
            with temporary_dir() as td:
                safe_copy(project2_bdist, os.path.join(td, os.path.basename(project2_bdist)))

                project1_pex = os.path.join(td, "project1.pex")
                result = run_pex_command(
                    ["-f", td, "-o", project1_pex, "{}[foo]".format(project1_dir)]
                )
                result.assert_success()

                output = subprocess.check_output(
                    [
                        project1_pex,
                        "-c",
                        "from project2 import my_module; my_module.do_something()",
                    ],
                    env=make_env(PEX_INTERPRETER="1"),
                )
                assert output.decode("utf-8").strip() == u"hello world!"



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import subprocess
import sys
from subprocess import CalledProcessError

import pytest

from pex.layout import Layout
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, Tuple

    import attr  # vendor:skip

    CreateColorsPex = Callable[[Iterable[str]], str]
    ExecuteColorsPex = Callable[[str, Dict[str, str]], Tuple[str, str]]
else:
    from pex.third_party import attr


@pytest.fixture
def create_colors_pex(tmpdir):
    # type: (Any) -> CreateColorsPex
    def create(extra_args):
        pex_file = os.path.join(str(tmpdir), "colors.pex")
        results = run_pex_command(["ansicolors==1.1.8", "-o", pex_file] + list(extra_args))
        results.assert_success()
        return pex_file

    return create


@pytest.fixture
def execute_colors_pex(tmpdir):
    # type: (Any) -> ExecuteColorsPex
    def execute(colors_pex, extra_env):
        pex_root = os.path.join(str(tmpdir), "pex_root")
        env = os.environ.copy()
        env.update(extra_env)
        env["PEX_ROOT"] = pex_root
        args = [colors_pex] if os.path.isfile(colors_pex) else [sys.executable, colors_pex]
        output = subprocess.check_output(
            args=args + ["-c", "import colors; print(colors.__file__)"], env=env
        )
        return output.strip().decode("utf-8"), pex_root

    return execute


@attr.s(frozen=True)
class ExecutionMode(object):
    extra_args = attr.ib()  # type: Iterable[str]
    isort_code_dir = attr.ib()  # type: Callable[[Layout.Value], str]
    venv_exception_expected = attr.ib()  # type: bool


def installed_wheels_or_deps(layout):
    # type: (Layout.Value) -> str
    return "{app_root}/.deps/" if layout == Layout.LOOSE else "{pex_root}/installed_wheels/"


@pytest.mark.parametrize(
    "execution_mode",
    [
        pytest.param(
            ExecutionMode(
                extra_args=[],
                isort_code_dir=installed_wheels_or_deps,
                venv_exception_expected=True,
            ),
            id="PEX",
        ),
        pytest.param(
            ExecutionMode(
                extra_args=["--include-tools"],
                isort_code_dir=installed_wheels_or_deps,
                venv_exception_expected=False,
            ),
            id="PEX --include-tools",
        ),
        pytest.param(
            ExecutionMode(
                extra_args=["--venv"],
                isort_code_dir=lambda _: "{pex_root}/venvs/",
                venv_exception_expected=False,
            ),
            id="VENV",
        ),
    ],
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
def test_execution_mode(
    create_colors_pex,  # type: CreateColorsPex
    execute_colors_pex,  # type: ExecuteColorsPex
    execution_mode,  # type: ExecutionMode
    layout,  # type: Layout.Value
):
    # type: (...) -> None
    pex_app = create_colors_pex(list(execution_mode.extra_args) + ["--layout", layout.value])

    output, pex_root = execute_colors_pex(pex_app, {})
    assert output.startswith(
        execution_mode.isort_code_dir(layout).format(app_root=pex_app, pex_root=pex_root),
    )

    if execution_mode.venv_exception_expected:
        with pytest.raises(CalledProcessError):
            execute_colors_pex(pex_app, {"PEX_VENV": "1"})
    else:
        output, pex_root = execute_colors_pex(pex_app, {"PEX_VENV": "1"})
        assert output.startswith(os.path.join(pex_root, "venvs"))



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pex.pip.version import PipVersion


def test_latest():
    # type: () -> None

    assert PipVersion.LATEST != PipVersion.VENDORED
    assert PipVersion.LATEST >= PipVersion.v23_1
    assert max(PipVersion.values(), key=lambda pv: pv.version) is PipVersion.LATEST



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
from textwrap import dedent

from pex import pex_warnings, third_party
from pex.atomic_directory import atomic_directory
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.pex import PEX
from pex.pex_bootstrapper import ensure_venv
from pex.pip.tool import Pip
from pex.pip.version import PipVersion, PipVersionValue
from pex.resolve.resolvers import Resolver
from pex.targets import LocalInterpreter, RequiresPythonError, Targets
from pex.third_party import isolated
from pex.typing import TYPE_CHECKING
from pex.util import named_temporary_file
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Callable, Dict, Iterator, Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def _pip_installation(
    version,  # type: PipVersionValue
    iter_distribution_locations,  # type: Callable[[], Iterator[str]]
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Pip
    pip_root = os.path.join(ENV.PEX_ROOT, "pip", str(version))
    path = os.path.join(pip_root, "pip.pex")
    pip_interpreter = interpreter or PythonInterpreter.get()
    pip_pex_path = os.path.join(path, isolated().pex_hash)
    with atomic_directory(pip_pex_path) as chroot:
        if not chroot.is_finalized():
            from pex.pex_builder import PEXBuilder

            isolated_pip_builder = PEXBuilder(path=chroot.work_dir)
            isolated_pip_builder.info.venv = True
            for dist_location in iter_distribution_locations():
                isolated_pip_builder.add_dist_location(dist=dist_location)
            with named_temporary_file(prefix="", suffix=".py", mode="w") as fp:
                fp.write(
                    dedent(
                        """\
                        import os
                        import runpy

                        patches_package = os.environ.pop({patches_package_env_var_name!r}, None)
                        if patches_package:
                            # Apply runtime patches to Pip to work around issues or else bend
                            # Pip to Pex's needs.
                            __import__(patches_package)

                        runpy.run_module(mod_name="pip", run_name="__main__", alter_sys=True)
                        """
                    ).format(patches_package_env_var_name=Pip._PATCHES_PACKAGE_ENV_VAR_NAME)
                )
                fp.close()
                isolated_pip_builder.set_executable(fp.name, "__pex_patched_pip__.py")
            isolated_pip_builder.freeze()
    pip_cache = os.path.join(pip_root, "pip_cache")
    pip_pex = ensure_venv(PEX(pip_pex_path, interpreter=pip_interpreter))
    return Pip(pip_pex=pip_pex, pip_cache=pip_cache)


def _vendored_installation(interpreter=None):
    # type: (Optional[PythonInterpreter]) -> Pip

    return _pip_installation(
        version=PipVersion.VENDORED,
        iter_distribution_locations=lambda: third_party.expose(("pip", "setuptools", "wheel")),
        interpreter=interpreter,
    )


def _resolved_installation(
    version,  # type: PipVersionValue
    resolver,  # type: Resolver
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> Pip
    if version is PipVersion.VENDORED:
        return _vendored_installation(interpreter=interpreter)

    def resolve_distribution_locations():
        for installed_distribution in resolver.resolve_requirements(
            requirements=version.requirements,
            targets=Targets(interpreters=(interpreter or PythonInterpreter.get(),)),
            pip_version=PipVersion.VENDORED,
        ).installed_distributions:
            yield installed_distribution.distribution.location

    return _pip_installation(
        version=version,
        iter_distribution_locations=resolve_distribution_locations,
        interpreter=interpreter,
    )


@attr.s(frozen=True)
class PipInstallation(object):
    interpreter = attr.ib()  # type: PythonInterpreter
    version = attr.ib()  # type: PipVersionValue

    def check_python_applies(self):
        # type: () -> None
        if not self.version.requires_python_applies(LocalInterpreter.create(self.interpreter)):
            raise RequiresPythonError(
                "The Pip requested was {pip_requirement} but it does not work with the interpreter "
                "selected which is {python_impl} {python_version} at {python_binary}. Pip "
                "{pip_version} requires Python {requires_python}.".format(
                    pip_requirement=self.version.requirement,
                    pip_version=self.version.value,
                    python_impl=self.interpreter.identity.interpreter,
                    python_version=self.interpreter.identity.version_str,
                    python_binary=self.interpreter.binary,
                    requires_python=self.version.requires_python,
                )
            )


def validate_targets(
    targets,  # type: Targets
    version,  # type: PipVersionValue
    context,  # type: str
):
    # type: (...) -> None
    all_targets = targets.unique_targets()
    invalid_targets = [
        target for target in all_targets if not version.requires_python_applies(target)
    ]
    if invalid_targets:
        raise RequiresPythonError(
            "The Pip requested for {context} was {pip_requirement} but it does not work with "
            "{quantifier} targets selected.\n"
            "\n"
            "Pip {pip_version} requires Python {requires_python} and the following {targets_do} "
            "not apply:\n"
            "{invalid_targets}"
            "".format(
                context=context,
                pip_requirement=version.requirement,
                quantifier="any of the"
                if len(invalid_targets) == len(all_targets)
                else "{invalid} out of the {total}".format(
                    invalid=len(invalid_targets), total=len(all_targets)
                ),
                pip_version=version.value,
                requires_python=version.requires_python,
                targets_do="target does" if len(invalid_targets) == 1 else "targets do",
                invalid_targets="\n".join(
                    "{index}. {target}".format(index=index, target=target)
                    for index, target in enumerate(invalid_targets, start=1)
                ),
            )
        )


def compatible_version(
    targets,  # type: Targets
    requested_version,  # type: PipVersionValue
    context,  # type: str
):
    # type: (...) -> PipVersionValue
    try:
        validate_targets(targets, requested_version, context)
        return requested_version
    except RequiresPythonError as e:
        remaining_versions = OrderedSet([requested_version] + list(PipVersion.values()))
        remaining_versions.discard(requested_version)
        for version in remaining_versions:
            try:
                validate_targets(targets, version, context)
                pex_warnings.warn(
                    "{err}\n" "\n" "Using Pip {version} instead.".format(err=e, version=version)
                )
                return version
            except RequiresPythonError:
                continue
    return PipVersion.v20_3_4_patched


_PIP = {}  # type: Dict[PipInstallation, Pip]


def get_pip(
    interpreter=None,
    version=PipVersion.VENDORED,  # type: PipVersionValue
    resolver=None,  # type: Optional[Resolver]
):
    # type: (...) -> Pip
    """Returns a lazily instantiated global Pip object that is safe for un-coordinated use."""
    installation = PipInstallation(
        interpreter=interpreter or PythonInterpreter.get(),
        version=version,
    )
    pip = _PIP.get(installation)
    if pip is None:
        installation.check_python_applies()
        if version is PipVersion.VENDORED:
            pip = _vendored_installation(interpreter=interpreter)
        else:
            if resolver is None:
                raise ValueError(
                    "A resolver is required to install {requirement}".format(
                        requirement=version.requirement
                    )
                )
            pip = _resolved_installation(
                version=version, resolver=resolver, interpreter=interpreter
            )
        _PIP[installation] = pip
    return pip



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import os.path
import subprocess
from textwrap import dedent

import pytest

from pex.common import safe_open
from pex.compatibility import PY2
from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(PY2, reason="Example code used to drive test is Python 3 only.")
def test_stderr_not_torn_down(tmpdir):
    # type: (Any) -> None

    exe = os.path.join(str(tmpdir), "exe")
    with safe_open(exe, "w") as fp:
        fp.write(
            dedent(
                """\
                import sys
                import logging
                import atexit
                import logging.handlers
                import queue
                import sys
                import faulthandler

                import absl.app as absl_app
                import absl.logging as absl_logging
                from absl.flags import FLAGS


                def run():
                    print("hello")
                    absl_logging.error("HELP ME")


                def init_sys_logging():
                    root_logger = logging.getLogger()

                    FLAGS.alsologtostderr = True

                    # No limit on queue size.
                    log_queue = queue.Queue(-1)
                    queue_forwarder = logging.handlers.QueueHandler(log_queue)
                    root_logger.addHandler(queue_forwarder)

                    queue_handlers = []

                    # If absl logging is enabled; re-parent it to the queue.
                    absl_handler = absl_logging.get_absl_handler()
                    if absl_handler in root_logger.handlers:
                        root_logger.handlers.remove(absl_handler)
                        queue_handlers.append(absl_handler)

                    queue_log_listener = logging.handlers.QueueListener(
                        log_queue, *queue_handlers, respect_handler_level=True
                    )
                    queue_log_listener.start()

                    atexit.register(queue_log_listener.stop)

                    FLAGS.mark_as_parsed()


                if __name__ == "__main__":
                    absl_logging.set_verbosity(0)
                    absl_logging.use_absl_handler()
                    absl_logging.get_absl_handler().use_absl_log_file()

                    faulthandler.enable()

                    init_sys_logging()

                    def run_wrapper(fn) -> int:
                        absl_app._run_main(lambda args: fn(), sys.argv)
                        return 0

                    sys.exit(run_wrapper(run))
                """
            )
        )
    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["absl-py==0.10.0", "--exe", exe, "-o", pex]).assert_success()
    process = subprocess.Popen(args=[pex], stderr=subprocess.PIPE)
    _, stderr = process.communicate()
    error = stderr.decode("utf-8")
    assert 0 == process.returncode
    assert "HELP ME" in error
